@@ -1,4 +1,4 @@
-/* Copyright (C) 1999, 2000, 2001, 2002, Massachusetts Institute of Technology.
+/* Copyright (C) 1999-2014 Massachusetts Institute of Technology.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -19,7 +19,7 @@
 #include <stdio.h>
 #include <math.h>
 
-#include "../config.h"
+#include "config.h"
 #include <mpiglue.h>
 #include <mpi_utils.h>
 #include <check.h>
@@ -43,6 +43,15 @@ extern void eigensolver_get_eigenvals_aux(evectmatrix Y, real *eigenvals,
 #define MIN2(a,b) ((a) < (b) ? (a) : (b))
 #define MAX2(a,b) ((a) > (b) ? (a) : (b))
 
+#if defined(SCALAR_LONG_DOUBLE_PREC)
+#  define fabs fabsl
+#  define cos cosl
+#  define sin sinl
+#  define sqrt sqrtl
+#  define atan atanl
+#  define atan2 atan2l
+#endif
+
 /* Evalutate op, and set t to the elapsed time (in seconds). */
 #define TIME_OP(t, op) { \
      mpiglue_clock_t xxx_time_op_start_time = MPIGLUE_CLOCK; \
@@ -54,7 +63,7 @@ extern void eigensolver_get_eigenvals_aux(evectmatrix Y, real *eigenvals,
 
 /**************************************************************************/
 
-#define EIGENSOLVER_MAX_ITERATIONS 10000
+#define EIGENSOLVER_MAX_ITERATIONS 100000
 #define FEEDBACK_TIME 4.0 /* elapsed time before we print progress feedback */
 
 /* Number of iterations after which to reset conjugate gradient
@@ -96,29 +105,52 @@ extern void eigensolver_get_eigenvals_aux(evectmatrix Y, real *eigenvals,
 
 typedef struct {
      sqmatrix YtAY, DtAD, symYtAD, YtY, DtD, symYtD, S1, S2, S3;
+     real lag, d_lag, trace_YtLY, trace_DtLD, trace_YtLD;
 } trace_func_data;
 
-static double trace_func(double theta, double *trace_deriv, void *data)
+static linmin_real trace_func(linmin_real theta, linmin_real *trace_deriv, void *data)
 {
-     double trace;
+     linmin_real trace;
      trace_func_data *d = (trace_func_data *) data;
+     linmin_real c = cos(theta), s = sin(theta);
 
      {
-	  double c = cos(theta), s = sin(theta);
 	  sqmatrix_copy(d->S1, d->YtY);
 	  sqmatrix_aApbB(c*c, d->S1, s*s, d->DtD);
 	  sqmatrix_ApaB(d->S1, 2*s*c, d->symYtD);
-	  sqmatrix_invert(d->S1, 1, d->S2);
+	  if (!sqmatrix_invert(d->S1, 1, d->S2)) {
+	       /* if c or s is small, we sometimes run into numerical
+		  difficulties, and it is better to use the Taylor expansion */
+	       if (c < 1e-4 * s && c != 0) {
+		    sqmatrix_copy(d->S1, d->DtD);
+		    CHECK(sqmatrix_invert(d->S1, 1, d->S2), "singular DtD!");
+		    sqmatrix_AeBC(d->S2, d->S1, 0, d->symYtD, 1);
+		    sqmatrix_AeBC(d->S3, d->S2, 0, d->S1, 1);
+		    sqmatrix_aApbB(1/(s*s), d->S1, -2*c/(s*s*s), d->S3);
+	       }
+	       else if (s < 1e-4 * c && s != 0) {
+		    sqmatrix_copy(d->S1, d->YtY);
+		    CHECK(sqmatrix_invert(d->S1, 1, d->S2), "singular DtD!");
+		    sqmatrix_AeBC(d->S2, d->S1, 0, d->symYtD, 1);
+		    sqmatrix_AeBC(d->S3, d->S2, 0, d->S1, 1);
+		    sqmatrix_aApbB(1/(c*c), d->S1, -2*s/(c*c*c), d->S3);
+	       }
+	       else {
+		    CHECK(0, "inexplicable singularity in linmin trace_func");
+	       }
+	  }
 	  
 	  sqmatrix_copy(d->S2, d->YtAY);
 	  sqmatrix_aApbB(c*c, d->S2, s*s, d->DtAD);
 	  sqmatrix_ApaB(d->S2, 2*s*c, d->symYtAD);
 	  
-	  trace = SCALAR_RE(sqmatrix_traceAtB(d->S2, d->S1));
+	  trace = SCALAR_RE(sqmatrix_traceAtB(d->S2, d->S1))
+	       + (c*c * d->trace_YtLY + s*s * d->trace_DtLD
+		  + 2*s*c * d->trace_YtLD) * (c * d->lag + s * d->d_lag);
      }
 
      if (trace_deriv) {
-	  double c2 = cos(2*theta), s2 = sin(2*theta);
+	  linmin_real c2 = cos(2*theta), s2 = sin(2*theta);
 	  
 	  sqmatrix_copy(d->S3, d->YtAY);
 	  sqmatrix_ApaB(d->S3, -1.0, d->DtAD);
@@ -135,6 +167,11 @@ static double trace_func(double theta, double *trace_deriv, void *data)
 	  
 	  *trace_deriv -= SCALAR_RE(sqmatrix_traceAtB(d->S2, d->S3));
 	  *trace_deriv *= 2;
+
+	  *trace_deriv += (-s2 * d->trace_YtLY + s2 * d->trace_DtLD
+		  + 2*c2 * d->trace_YtLD) * (c * d->lag + s * d->d_lag);
+	  *trace_deriv += (c*c * d->trace_YtLY + s*s * d->trace_DtLD
+		  + 2*s*c * d->trace_YtLD) * (-s * d->lag + c * d->d_lag);
      }
 
      return trace;
@@ -144,23 +181,38 @@ static double trace_func(double theta, double *trace_deriv, void *data)
 
 #define EIG_HISTORY_SIZE 5
 
-void eigensolver(evectmatrix Y, real *eigenvals,
-                 evectoperator A, void *Adata,
-                 evectpreconditioner K, void *Kdata,
-                 evectconstraint constraint, void *constraint_data,
-                 evectmatrix Work[], int nWork,
-                 real tolerance, int *num_iterations,
-                 int flags)
+/* find eigenvectors Y of A by minimizing Rayleigh quotient
+
+        tr [ Yt A Y / (Yt Y) ] + lag * tr [ Yt L Y ]
+
+   where lag is a Lagrange multiplier and L is a Hermitian operator
+   implementing some constraint tr [ Yt L Y ] = 0 on the eigenvectors
+   (if L is not NULL).
+
+   Constraints that commute with A (and L) are specified via the
+   "constraint" argument, which gives the projection operator for
+   the constraint(s).
+*/
+
+void eigensolver_lagrange(evectmatrix Y, real *eigenvals,
+			  evectoperator A, void *Adata,
+			  evectpreconditioner K, void *Kdata,
+			  evectconstraint constraint, void *constraint_data,
+			  evectoperator L, void *Ldata, real *lag,
+			  evectmatrix Work[], int nWork,
+			  real tolerance, int *num_iterations,
+			  int flags)
 {
      real convergence_history[EIG_HISTORY_SIZE];
      evectmatrix G, D, X, prev_G;
+     real g_lag = 0, d_lag = 0, prev_g_lag = 0;
      short usingConjugateGradient = 0, use_polak_ribiere = 0,
 	   use_linmin = 1;
      real E, prev_E = 0.0;
      real d_scale = 1.0;
      real traceGtX, prev_traceGtX = 0.0;
      real theta, prev_theta = 0.5;
-     int i, iteration = 0;
+     int i, iteration = 0, num_emergency_restarts = 0;
      mpiglue_clock_t prev_feedback_time;
      real time_AZ, time_KZ=0, time_ZtZ, time_ZtW, time_ZS, time_linmin=0;
      real linmin_improvement = 0;
@@ -218,9 +270,12 @@ void eigensolver(evectmatrix Y, real *eigenvals,
      tfd.YtY = YtY; tfd.DtD = DtD; tfd.symYtD = symYtD;
      tfd.S1 = YtAYU; tfd.S2 = S2; tfd.S3 = S3;
 
+ restartY:
+
      if (flags & EIGS_ORTHONORMALIZE_FIRST_STEP) {
 	  evectmatrix_XtX(U, Y, S2);
-	  sqmatrix_invert(U, 1, S2);
+	  sqmatrix_assert_hermitian(U);
+	  CHECK(sqmatrix_invert(U, 1, S2), "non-independent initial Y");
 	  sqmatrix_sqrt(S1, U, S2); /* S1 = 1/sqrt(Yt*Y) */
 	  evectmatrix_XeYS(G, Y, S1, 1); /* G = orthonormalize Y */
 	  evectmatrix_copy(Y, G);
@@ -242,13 +297,26 @@ void eigensolver(evectmatrix Y, real *eigenvals,
 	       use_linmin = 0;
 
 	  TIME_OP(time_ZtZ, evectmatrix_XtX(YtY, Y, S2));
+	  sqmatrix_assert_hermitian(YtY);
 
 	  y_norm = sqrt(SCALAR_RE(sqmatrix_trace(YtY)) / Y.p);
 	  blasglue_rscal(Y.p * Y.n, 1/y_norm, Y.data, 1);
 	  blasglue_rscal(Y.p * Y.p, 1/(y_norm*y_norm), YtY.data, 1);
 
 	  sqmatrix_copy(U, YtY);
-	  sqmatrix_invert(U, 1, S2);
+	  if (!sqmatrix_invert(U, 1, S2)) { /* non-independent Y columns */
+	       /* emergency restart with random Y */
+	       CHECK(iteration + 10 * ++num_emergency_restarts
+		     < EIGENSOLVER_MAX_ITERATIONS, 
+		     "too many emergency restarts");
+	       mpi_one_printf("    emergency randomization of Y on iter. %d\n",
+			      iteration);
+	       for (i = 0; i < Y.p * Y.n; ++i)
+		    ASSIGN_SCALAR(Y.data[i],
+				  rand() * 1.0 / RAND_MAX - 0.5,
+				  rand() * 1.0 / RAND_MAX - 0.5);
+	       goto restartY;
+	  }
 
 	  /* If trace(1/YtY) gets big, it means that the columns
 	     of Y are becoming nearly parallel.  This sometimes happens,
@@ -272,12 +340,18 @@ void eigensolver(evectmatrix Y, real *eigenvals,
 		    blasglue_rscal(Y.p * Y.n, 1/y_norm, Y.data, 1);
 		    blasglue_rscal(Y.p * Y.p, 1/(y_norm*y_norm), YtY.data, 1);
 		    sqmatrix_copy(U, YtY);
-		    sqmatrix_invert(U, 1, S2);
+		    CHECK(sqmatrix_invert(U, 1, S2),
+			  "non-independent Y after re-orthogonalization");
 	       }
 	  }
 
 	  TIME_OP(time_AZ, A(Y, X, Adata, 1, G)); /* X = AY; G is scratch */
 
+#ifdef DEBUG
+	  evectmatrix_XtY(S1, Y, X, S2);
+	  sqmatrix_assert_hermitian(S1);
+#endif
+
 	  /* G = AYU; note that U is Hermitian: */
 	  TIME_OP(time_ZS, evectmatrix_XeYS(G, X, U, 1));
 
@@ -286,6 +360,13 @@ void eigensolver(evectmatrix Y, real *eigenvals,
 	  CHECK(!BADNUM(E), "crazy number detected in trace!!\n");
 	  mpi_assert_equal(E);
 
+	  if (L) {
+	       /* X = LY, no scratch */
+	       L(Y, X, Ldata, 1, X);
+	       g_lag = tfd.trace_YtLY = SCALAR_RE(evectmatrix_traceXtY(Y, X));
+	       E += *lag * g_lag;
+	  }
+
 	  convergence_history[iteration % EIG_HISTORY_SIZE] =
 	       200.0 * fabs(E - prev_E) / (fabs(E) + fabs(prev_E));
 	  
@@ -293,9 +374,9 @@ void eigensolver(evectmatrix Y, real *eigenvals,
 	      ((flags & EIGS_VERBOSE) ||
 	       MPIGLUE_CLOCK_DIFF(MPIGLUE_CLOCK, prev_feedback_time)
 	       > FEEDBACK_TIME)) {
-               printf("    iteration %4d: "
-                      "trace = %0.16g (%g%% change)\n", iteration, E,
-		      convergence_history[iteration % EIG_HISTORY_SIZE]);
+               mpi_one_printf("    iteration %4d: "
+                      "trace = %0.16g (%g%% change)\n", iteration, (double)E,
+	           (double)convergence_history[iteration % EIG_HISTORY_SIZE]);
 	       if (flags & EIGS_VERBOSE)
 		    debug_output_malloc_count();
 	       fflush(stdout); /* make sure output appears */
@@ -310,6 +391,10 @@ void eigensolver(evectmatrix Y, real *eigenvals,
 	  sqmatrix_AeBC(S1, U, 0, YtAYU, 0);
 	  evectmatrix_XpaYS(G, -1.0, Y, S1, 1);
 
+	  if (L) { /* include Lagrange gradient; note X = LY from above */
+	       evectmatrix_aXpbY(1.0, G, *lag, X);
+	  }
+
 	  /* set X = precondition(G): */
 	  if (K != NULL) {
 	       TIME_OP(time_KZ, K(G, X, Kdata, Y, NULL, YtY));
@@ -337,7 +422,9 @@ void eigensolver(evectmatrix Y, real *eigenvals,
 	     any computations that we need with G.  (Yes, we're
 	     playing tricksy games here, but isn't it fun?) */
 
-	  mpi_assert_equal(traceGtX = SCALAR_RE(evectmatrix_traceXtY(G, X)));
+	  mpi_assert_equal(traceGtX = 
+			   SCALAR_RE(evectmatrix_traceXtY(G, X))
+			   + g_lag * g_lag);
 	  if (usingConjugateGradient) {
                if (use_polak_ribiere) {
                     /* assign G = G - prev_G and copy prev_G = G in the
@@ -349,6 +436,9 @@ void eigensolver(evectmatrix Y, real *eigenvals,
                          prev_G.data[i] = g;
                     }
                     gamma_numerator = SCALAR_RE(evectmatrix_traceXtY(G, X));
+
+		    { real g = g_lag; g_lag -= prev_g_lag; prev_g_lag = g; }
+		    gamma_numerator += g_lag * prev_g_lag;
                }
                else /* otherwise, use Fletcher-Reeves (ignore prev_G) */
                     gamma_numerator = traceGtX;
@@ -378,7 +468,8 @@ void eigensolver(evectmatrix Y, real *eigenvals,
 		  S3 = (YtKG)t / (YtKY).  Recall that, at this point,
 		  X holds KG and G holds KY.  K is assumed Hermitian. */
 	       evectmatrix_XtY(S1, Y, G, S2);
-	       sqmatrix_invert(S1, 0, S2);  /* S1 = 1 / (YtKY) */
+	       CHECK(sqmatrix_invert(S1, 0, S2),
+		     "singular YtKY");  /* S1 = 1 / (YtKY) */
 	       evectmatrix_XtY(S2, X, Y, S3);  /* S2 = GtKY = (YtKG)t */
 	       sqmatrix_AeBC(S3, S2, 0 , S1, 1);
 	       evectmatrix_XpaYS(X, -1.0, G, S3, 1);
@@ -429,6 +520,7 @@ void eigensolver(evectmatrix Y, real *eigenvals,
 
 	       mpi_assert_equal(gamma * d_scale);
                evectmatrix_aXpbY(gamma * d_scale, D, 1.0, X);
+	       d_lag = gamma * d_scale * d_lag + g_lag;
           }
 
 	  d_scale = 1.0;
@@ -458,11 +550,19 @@ void eigensolver(evectmatrix Y, real *eigenvals,
 	       evectmatrix_aXpbY(1.0, Y, t / d_norm, D);
 
 	       evectmatrix_XtX(U, Y, S2);
-	       sqmatrix_invert(U, 1, S2);  /* U = 1 / (Yt Y) */
+	       CHECK(sqmatrix_invert(U, 1, S2),
+		     "singular YtY");  /* U = 1 / (Yt Y) */
 	       A(Y, G, Adata, 1, X); /* G = AY; X is scratch */
 	       evectmatrix_XtY(S1, Y, G, S2);  /* S1 = Yt A Y */
-	       
+
 	       E2 = SCALAR_RE(sqmatrix_traceAtB(S1, U));
+
+	       if (L) {
+		    *lag += (t / d_norm) * d_lag;
+		    L(Y, X, Ldata, 1, X);
+		    E2 += *lag * SCALAR_RE(evectmatrix_traceXtY(Y, X));
+	       }
+	       
 	       mpi_assert_equal(E2);
 
 	       /* Get finite-difference approximation for the 2nd derivative
@@ -489,6 +589,7 @@ void eigensolver(evectmatrix Y, real *eigenvals,
 					     "line minimization\n");
 			 use_linmin = 1;
 			 evectmatrix_aXpbY(1.0, Y, -t / d_norm, D);
+			 if (L) *lag -= (t / d_norm) * d_lag;
 			 prev_theta = atan(prev_theta); /* convert to angle */
 			 /* don't do this again: */
 			 flags |= EIGS_FORCE_EXACT_LINMIN;
@@ -497,6 +598,7 @@ void eigensolver(evectmatrix Y, real *eigenvals,
 	       else {
 		    /* Shift Y by theta, hopefully minimizing the trace: */
 		    evectmatrix_aXpbY(1.0, Y, (theta - t) / d_norm, D);
+		    if (L) *lag += ((theta - t) / d_norm) * d_lag;
 	       }
 	  }
 	  if (use_linmin) {
@@ -509,6 +611,8 @@ void eigensolver(evectmatrix Y, real *eigenvals,
 	       A(D, G, Adata, 0, X); /* G = A D; X is scratch */
 	       evectmatrix_XtX(DtD, D, S2);
 	       evectmatrix_XtY(DtAD, D, G, S2);
+	       sqmatrix_assert_hermitian(DtD);
+	       sqmatrix_assert_hermitian(DtAD);
 	       
 	       evectmatrix_XtY(S1, Y, D, S2);
 	       sqmatrix_symmetrize(symYtD, S1);
@@ -528,6 +632,24 @@ void eigensolver(evectmatrix Y, real *eigenvals,
 			    SCALAR_RE(sqmatrix_traceAtB(YtAYU, S1)) -
 			    4.0 * SCALAR_RE(sqmatrix_traceAtB(U, S3)));
 	       
+	       if (L) {
+		    d_lag *= 1/d_scale;
+		    tfd.d_lag = d_lag;
+		    tfd.lag = *lag;
+		    /* note: tfd.trace_YtLY was set above */
+		    L(D, X, Ldata, 0, X);
+		    tfd.trace_DtLD = SCALAR_RE(evectmatrix_traceXtY(D, X));
+		    tfd.trace_YtLD = SCALAR_RE(evectmatrix_traceXtY(Y, X));
+		    dE += tfd.lag * 2.0 * tfd.trace_YtLD
+			 + tfd.d_lag * tfd.trace_YtLY;
+		    d2E += tfd.lag * 2.0 * tfd.trace_DtLD
+			 + tfd.d_lag * 4.0 * tfd.trace_YtLD;
+	       }
+	       else {
+		    tfd.d_lag = tfd.lag = tfd.trace_YtLY = 
+			 tfd.trace_DtLD = tfd.trace_YtLD = 0;
+	       }
+
 	       /* this is just Newton-Raphson to find a root of
 		  the first derivative: */
 	       theta = -dE/d2E;
@@ -540,21 +662,22 @@ void eigensolver(evectmatrix Y, real *eigenvals,
 	       else if (-0.5*dE*theta > 2.0 * fabs(E-prev_E)) {
 		    if (flags & EIGS_VERBOSE)
 			 mpi_one_printf("    large trace change predicted "
-					"(%g%%)\n", -0.5*dE*theta/E * 100.0);
+					"(%g%%)\n", (double) (-0.5*dE*theta/E * 100.0));
 	       }
 	       if (fabs(theta) >= K_PI) {
 		    if (flags & EIGS_VERBOSE)
-			 mpi_one_printf("    large theta (%g)\n", theta);
+			 mpi_one_printf("    large theta (%g)\n", (double)theta);
 		    theta = dE > 0 ? -fabs(prev_theta) : fabs(prev_theta);
 	       }
 
 	       /* Set S1 to YtAYU * YtY = YtAY for use in linmin.
 		  (tfd.YtAY == S1). */
 	       sqmatrix_AeBC(S1, YtAYU, 0, YtY, 1);
+	       sqmatrix_assert_hermitian(S1);
 
 	       mpi_assert_equal(theta);
 	       {
-		    double new_E, new_dE;
+		    linmin_real new_E, new_dE;
 		    TIME_OP(time_linmin,
 			    theta = linmin(&new_E, &new_dE, theta, E, dE,
 					   0.1, MIN2(tolerance, 1e-6), 1e-14,
@@ -569,6 +692,7 @@ void eigensolver(evectmatrix Y, real *eigenvals,
 
 	       /* Shift Y to new location minimizing the trace along D: */
 	       evectmatrix_aXpbY(cos(theta), Y, sin(theta), D);
+	       if (L) *lag = *lag * cos(theta) + d_lag * sin(theta);
 	  }
 
 	  /* In exact arithmetic, we don't need to do this, but in practice
@@ -608,7 +732,7 @@ void eigensolver(evectmatrix Y, real *eigenvals,
 		    if ((flags & EIGS_VERBOSE) && use_linmin)
 			 mpi_one_printf("    switching to approximate "
 				"line minimization (decrease time by %g%%)\n",
-				(t_exact - t_approx) * 100.0 / t_exact);
+			    (double) ((t_exact - t_approx) * 100.0 / t_exact));
 		    use_linmin = 0;
 	       }
 	       else if (!(flags & EIGS_FORCE_APPROX_LINMIN)) {
@@ -627,7 +751,7 @@ void eigensolver(evectmatrix Y, real *eigenvals,
            " iterations");
 
      evectmatrix_XtX(U, Y, S2);
-     sqmatrix_invert(U, 1, S2);
+     CHECK(sqmatrix_invert(U, 1, S2), "singular YtY at end");
      eigensolver_get_eigenvals_aux(Y, eigenvals, A, Adata,
 				   X, G, U, S1, S2);
 
@@ -644,3 +768,17 @@ void eigensolver(evectmatrix Y, real *eigenvals,
      destroy_sqmatrix(DtAD);
      destroy_sqmatrix(YtAYU);
 }
+
+void eigensolver(evectmatrix Y, real *eigenvals,
+		 evectoperator A, void *Adata,
+		 evectpreconditioner K, void *Kdata,
+		 evectconstraint constraint, void *constraint_data,
+		 evectmatrix Work[], int nWork,
+		 real tolerance, int *num_iterations,
+		 int flags)
+{
+     eigensolver_lagrange(Y, eigenvals, A, Adata, K, Kdata,
+			  constraint, constraint_data,
+			  0, 0, 0,
+			  Work, nWork, tolerance, num_iterations, flags);
+}
