@@ -1,4 +1,4 @@
-/* Copyright (C) 1999, 2000, 2001, 2002, Massachusetts Institute of Technology.
+/* Copyright (C) 1999-2014 Massachusetts Institute of Technology.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -24,15 +24,16 @@
 #include <stdio.h>
 #include <string.h>
 
-#include "../config.h"
+#include "config.h"
 
 #include <check.h>
 #include <scalar.h>
-#include <mpiglue.h>
 #include <mpi_utils.h>
 
 #include "matrixio.h"
 
+#include <mpiglue.h>
+
 /*****************************************************************************/
 
 /* If we have the H5Pset_fapl_mpio function (which is available if HDF5 was
@@ -78,66 +79,70 @@
    can *not* be attached to files, in which case we'll write/read it
    as an ordinary dataset.  Ugh. */
 
-static void write_attr(matrixio_id id, matrixio_id type_id,
-		       matrixio_id space_id, const char *name, void *val)
+static void write_attr(matrixio_id id, matrixio_id_ type_id,
+		       matrixio_id_ space_id,
+		       const char *name, const void *val)
 {
 #if defined(HAVE_HDF5)
      hid_t attr_id;
 
-     if (!mpi_is_master())
+#ifndef HAVE_H5PSET_FAPL_MPIO
+     if (!mpi_is_master() && id.parallel)
 	  return; /* only one process should add attributes */
+#else
+     /* otherwise, the operations must be performed collectively */
+#endif
      
-     if (H5I_FILE == H5Iget_type(id)) {
-          attr_id = H5Dcreate(id, name, type_id, space_id, H5P_DEFAULT);
-          CHECK(id >= 0, "error creating HDF attr");
+     if (H5I_FILE == H5Iget_type(id.id)) {
+          attr_id = H5Dcreate(id.id, name, type_id, space_id, H5P_DEFAULT);
+          CHECK(attr_id >= 0, "error creating HDF attr");
           H5Dwrite(attr_id, type_id, H5S_ALL, H5S_ALL, H5P_DEFAULT, val);
           H5Dclose(attr_id);
      }
      else {
-	  attr_id = H5Acreate(id, name, type_id, space_id, H5P_DEFAULT);
-	  CHECK(id >= 0, "error creating HDF attr");
+	  attr_id = H5Acreate(id.id, name, type_id, space_id, H5P_DEFAULT);
+	  CHECK(attr_id >= 0, "error creating HDF attr");
 	  H5Awrite(attr_id, type_id, val);
 	  H5Aclose(attr_id);
      }
 #endif
 }
 
-static matrixio_id open_attr(matrixio_id id, matrixio_id *type_id,
-			     matrixio_id *space_id, const char *name)
+static matrixio_id open_attr(matrixio_id id, matrixio_id_ *type_id,
+			     matrixio_id_ *space_id, const char *name)
 {
+     matrixio_id attr_id;
+     attr_id.parallel = id.parallel;
+     attr_id.id = -1;
 #if defined(HAVE_HDF5)
-     hid_t attr_id;
-
-     if (H5I_FILE == H5Iget_type(id)) {
-          SUPPRESS_HDF5_ERRORS(attr_id = H5Dopen(id, name));
-	  if (attr_id >= 0) {
-	       *type_id = H5Dget_type(attr_id);
-	       *space_id = H5Dget_space(attr_id);
+     if (H5I_FILE == H5Iget_type(id.id)) {
+          SUPPRESS_HDF5_ERRORS(attr_id.id = H5Dopen(id.id, name));
+	  if (attr_id.id >= 0) {
+	       *type_id = H5Dget_type(attr_id.id);
+	       *space_id = H5Dget_space(attr_id.id);
 	  }
      }
      else {
-          SUPPRESS_HDF5_ERRORS(attr_id = H5Aopen_name(id, name));
-	  if (attr_id >= 0) {
-	       *type_id = H5Aget_type(attr_id);
-	       *space_id = H5Aget_space(attr_id);
+          SUPPRESS_HDF5_ERRORS(attr_id.id = H5Aopen_name(id.id, name));
+	  if (attr_id.id >= 0) {
+	       *type_id = H5Aget_type(attr_id.id);
+	       *space_id = H5Aget_space(attr_id.id);
 	  }
      }
 
-     return attr_id;
-#else
-     return -1;
 #endif
+     return attr_id;
 }
 
 static void read_attr(matrixio_id id, matrixio_id attr_id,
-		      matrixio_id mem_type_id, void *val)
+		      matrixio_id_ mem_type_id, void *val)
 {
 #if defined(HAVE_HDF5)
-     if (H5I_FILE == H5Iget_type(id)) {
-	  H5Dread(attr_id, mem_type_id, H5S_ALL, H5S_ALL, H5P_DEFAULT, val);
+     if (H5I_FILE == H5Iget_type(id.id)) {
+	  H5Dread(attr_id.id, mem_type_id, H5S_ALL, H5S_ALL, H5P_DEFAULT, val);
      }
      else {
-	  H5Aread(attr_id, mem_type_id, val);
+	  H5Aread(attr_id.id, mem_type_id, val);
      }
 #endif
 }
@@ -145,11 +150,11 @@ static void read_attr(matrixio_id id, matrixio_id attr_id,
 static void close_attr(matrixio_id id, matrixio_id attr_id)
 {
 #if defined(HAVE_HDF5)
-     if (H5I_FILE == H5Iget_type(id)) {
-          H5Dclose(attr_id);
+     if (H5I_FILE == H5Iget_type(id.id)) {
+          H5Dclose(attr_id.id);
      }
      else {
-          H5Aclose(attr_id);
+          H5Aclose(attr_id.id);
      }
 #endif
 }
@@ -169,7 +174,7 @@ void matrixio_write_string_attr(matrixio_id id, const char *name,
      type_id = H5Tcopy(H5T_C_S1);
      H5Tset_size(type_id, strlen(val) + 1);
      space_id = H5Screate(H5S_SCALAR);
-     write_attr(id, type_id, space_id, name, (void*) val);
+     write_attr(id, type_id, space_id, name, val);
      H5Sclose(space_id);
      H5Tclose(type_id);
 #endif
@@ -187,8 +192,10 @@ void matrixio_write_data_attr(matrixio_id id, const char *name,
      if (!val || !name || !name[0] || rank < 0 || !dims)
 	  return; /* don't try to create empty attributes */
      
-#ifdef SCALAR_SINGLE_PREC
+#if defined(SCALAR_SINGLE_PREC)
      type_id = H5T_NATIVE_FLOAT;
+#elif defined(SCALAR_LONG_DOUBLE_PREC)
+     type_id = H5T_NATIVE_LDOUBLE;
 #else
      type_id = H5T_NATIVE_DOUBLE;
 #endif
@@ -204,7 +211,7 @@ void matrixio_write_data_attr(matrixio_id id, const char *name,
 	  space_id = H5Screate(H5S_SCALAR);
      }
 
-     write_attr(id, type_id, space_id, name, (void*) val);
+     write_attr(id, type_id, space_id, name, val);
      H5Sclose(space_id);
 #endif
 }
@@ -212,7 +219,7 @@ void matrixio_write_data_attr(matrixio_id id, const char *name,
 char *matrixio_read_string_attr(matrixio_id id, const char *name)
 {
 #if defined(HAVE_HDF5)
-     hid_t attr_id;
+     matrixio_id attr_id;
      hid_t type_id;
      hid_t space_id;
      int len;
@@ -222,7 +229,7 @@ char *matrixio_read_string_attr(matrixio_id id, const char *name)
 	  return NULL; /* don't try to read empty-named attributes */
      
      attr_id = open_attr(id, &type_id, &space_id, name);
-     if (attr_id < 0)
+     if (attr_id.id < 0)
 	  return NULL;
 
      if (H5Sget_simple_extent_npoints(space_id) == 1) {
@@ -233,7 +240,7 @@ char *matrixio_read_string_attr(matrixio_id id, const char *name)
 	  H5Tset_size(type_id, len);
 	  
 	  CHK_MALLOC(s, char, len);
-	  read_attr(id, attr_id, type_id, (void*) s);
+	  read_attr(id, attr_id, type_id, s);
      }
 
      H5Tclose(type_id);
@@ -250,20 +257,23 @@ real *matrixio_read_data_attr(matrixio_id id, const char *name,
 			      int *rank, int max_rank, int *dims)
 {
 #if defined(HAVE_HDF5)
-     hid_t attr_id, type_id, mem_type_id, space_id;
+     matrixio_id attr_id;
+     hid_t type_id, mem_type_id, space_id;
      real *d = NULL;
 
      if (!name || !name[0] || max_rank < 0 || !dims)
 	  return NULL; /* don't try to create empty attributes */
      
-#ifdef SCALAR_SINGLE_PREC
+#if defined(SCALAR_SINGLE_PREC)
      mem_type_id = H5T_NATIVE_FLOAT;
+#elif defined(SCALAR_LONG_DOUBLE_PREC)
+     mem_type_id = H5T_NATIVE_LDOUBLE;
 #else
      mem_type_id = H5T_NATIVE_DOUBLE;
 #endif
 
      attr_id = open_attr(id, &type_id, &space_id, name);
-     if (attr_id < 0)
+     if (attr_id.id < 0)
           return NULL;
 
      *rank = H5Sget_simple_extent_ndims(space_id);
@@ -280,7 +290,7 @@ real *matrixio_read_data_attr(matrixio_id id, const char *name,
 	       free(space_dims);
 	  }
 	  CHK_MALLOC(d, real, H5Sget_simple_extent_npoints(space_id));
-          read_attr(id, attr_id, mem_type_id, (void*) d);
+          read_attr(id, attr_id, mem_type_id, d);
      }
 
      H5Tclose(type_id);
@@ -318,9 +328,11 @@ static char *add_fname_suffix(const char *fname)
 
 /*****************************************************************************/
 
+#ifndef HAVE_H5PSET_FAPL_MPIO
 static int matrixio_critical_section_tag = 0;
+#endif
 
-matrixio_id matrixio_create(const char *fname)
+static matrixio_id matrixio_create_(const char *fname, int parallel)
 {
 #if defined(HAVE_HDF5)
      char *new_fname;
@@ -330,22 +342,25 @@ matrixio_id matrixio_create(const char *fname)
      access_props = H5Pcreate (H5P_FILE_ACCESS);
      
 #  if defined(HAVE_MPI) && defined(HAVE_H5PSET_FAPL_MPIO)
-     H5Pset_fapl_mpio(access_props, MPI_COMM_WORLD, MPI_INFO_NULL);
+     if (parallel)
+	  CHECK(H5Pset_fapl_mpio(access_props, MPI_COMM_WORLD, MPI_INFO_NULL)
+		>= 0, "error initializing MPI file access");
 #  endif
 
      new_fname = add_fname_suffix(fname);
 
 #  ifdef HAVE_H5PSET_FAPL_MPIO
-     id = H5Fcreate(new_fname, H5F_ACC_TRUNC, H5P_DEFAULT, access_props);
+     id.id = H5Fcreate(new_fname, H5F_ACC_TRUNC, H5P_DEFAULT, access_props);
 #  else
-     mpi_begin_critical_section(matrixio_critical_section_tag);
-     if (mpi_is_master())
-	  id = H5Fcreate(new_fname, H5F_ACC_TRUNC, H5P_DEFAULT, access_props);
+     if (parallel) mpi_begin_critical_section(matrixio_critical_section_tag);
+     if (mpi_is_master() || !parallel)
+	  id.id = H5Fcreate(new_fname, H5F_ACC_TRUNC,H5P_DEFAULT,access_props);
      else
-	  id = H5Fopen(new_fname, H5F_ACC_RDWR, access_props);
+	  id.id = H5Fopen(new_fname, H5F_ACC_RDWR, access_props);
 #  endif
+     id.parallel = parallel;
 
-     CHECK(id >= 0, "error creating HDF output file");
+     CHECK(id.id >= 0, "error creating HDF output file");
 
      free(new_fname);
 
@@ -356,11 +371,22 @@ matrixio_id matrixio_create(const char *fname)
      mpi_one_fprintf(stderr,
 		     "matrixio: cannot output \"%s\" (compiled without HDF)\n",
 		     fname);
-     return 0;
+     {
+	  matrixio_id id = {0,0};
+	  return id;
+     }
 #endif
 }
 
-matrixio_id matrixio_open(const char *fname, int read_only)
+matrixio_id matrixio_create(const char *fname) {
+     return matrixio_create_(fname, 1);
+}
+
+matrixio_id matrixio_create_serial(const char *fname) {
+     return matrixio_create_(fname, 0);
+}
+
+static matrixio_id matrixio_open_(const char *fname, int read_only, int parallel)
 {
 #if defined(HAVE_HDF5)
      char *new_fname;
@@ -370,18 +396,20 @@ matrixio_id matrixio_open(const char *fname, int read_only)
      access_props = H5Pcreate (H5P_FILE_ACCESS);
      
 #  if defined(HAVE_MPI) && defined(HAVE_H5PSET_FAPL_MPIO)
-     H5Pset_fapl_mpio(access_props, MPI_COMM_WORLD, MPI_INFO_NULL);
+     if (parallel)
+	  H5Pset_fapl_mpio(access_props, MPI_COMM_WORLD, MPI_INFO_NULL);
 #  endif
 
      new_fname = add_fname_suffix(fname);
 
-     IF_EXCLUSIVE(mpi_begin_critical_section(matrixio_critical_section_tag),0);
+     IF_EXCLUSIVE(if (parallel) mpi_begin_critical_section(matrixio_critical_section_tag),0);
 
      if (read_only)
-	  id = H5Fopen(new_fname, H5F_ACC_RDONLY, access_props);
+	  id.id = H5Fopen(new_fname, H5F_ACC_RDONLY, access_props);
      else
-	  id = H5Fopen(new_fname, H5F_ACC_RDWR, access_props);
-     CHECK(id >= 0, "error opening HDF input file");
+	  id.id = H5Fopen(new_fname, H5F_ACC_RDWR, access_props);
+     id.parallel = parallel;
+     CHECK(id.id >= 0, "error opening HDF input file");
 
      free(new_fname);
 
@@ -390,55 +418,70 @@ matrixio_id matrixio_open(const char *fname, int read_only)
      return id;
 #else
      CHECK(0, "no matrixio implementation is linked");
-     return 0;
+     {
+	  matrixio_id id = {0,0};
+	  return id;
+     }
 #endif
 }
 
 void matrixio_close(matrixio_id id)
 {
 #if defined(HAVE_HDF5)
-     CHECK(H5Fclose(id) >= 0, "error closing HDF file");
-     IF_EXCLUSIVE(mpi_end_critical_section(matrixio_critical_section_tag++),0);
+     CHECK(H5Fclose(id.id) >= 0, "error closing HDF file");
+     IF_EXCLUSIVE(if (id.parallel) mpi_end_critical_section(matrixio_critical_section_tag++),0);
 #endif
 }
 
+matrixio_id matrixio_open(const char *fname, int read_only) {
+     return matrixio_open_(fname, read_only, 1);
+}
+
+matrixio_id matrixio_open_serial(const char *fname, int read_only) {
+     return matrixio_open_(fname, read_only, 0);
+}
+
 /*****************************************************************************/
 
 matrixio_id matrixio_create_sub(matrixio_id id, 
 				const char *name, const char *description)
 {
-#if defined(HAVE_HDF5)
      matrixio_id sub_id;
+     sub_id.id = 0;
+     sub_id.parallel = id.parallel;
+#if defined(HAVE_HDF5)
 
+#  ifdef HAVE_H5PSET_FAPL_MPIO /* H5Gcreate is collective */
+     sub_id.id = H5Gcreate(id.id, name, 0 /* ==> default size */ );
+     matrixio_write_string_attr(sub_id, "description", description);
+#  else
      /* when running a parallel job, only the master process creates the
 	group.  It flushes the group to disk and then the other processes
 	open the group.  Is this the right thing to do, or is the
         H5Gcreate function parallel-aware? */
 
-     if (mpi_is_master()) {
-	  sub_id = H5Gcreate(id, name, 0 /* ==> default size */ );
+     if (mpi_is_master() || !id.parallel) {
+	  sub_id.id = H5Gcreate(id.id, name, 0 /* ==> default size */ );
 	  matrixio_write_string_attr(sub_id, "description", description);
 	  
-	  H5Fflush(sub_id, H5F_SCOPE_GLOBAL);
+	  H5Fflush(sub_id.id, H5F_SCOPE_GLOBAL);
 
-	  IF_EXCLUSIVE(0,MPI_Barrier(MPI_COMM_WORLD));
+	  IF_EXCLUSIVE(0,if (id.parallel) MPI_Barrier(MPI_COMM_WORLD));
      }
      else {
-	  IF_EXCLUSIVE(0,MPI_Barrier(MPI_COMM_WORLD));
+	  IF_EXCLUSIVE(0,if (id.parallel) MPI_Barrier(MPI_COMM_WORLD));
 
-	  sub_id = H5Gopen(id, name);
+	  sub_id.id = H5Gopen(id.id, name);
      }
-
-     return sub_id;
-#else
-     return 0;
+#  endif
 #endif
+     return sub_id;
 }
 
 void matrixio_close_sub(matrixio_id id)
 {
 #if defined(HAVE_HDF5)
-     CHECK(H5Gclose(id) >= 0, "error closing HDF group");
+     CHECK(H5Gclose(id.id) >= 0, "error closing HDF group");
 #endif
 }
 
@@ -448,16 +491,18 @@ matrixio_id matrixio_open_dataset(matrixio_id id,
 				  const char *name,
 				  int rank, const int *dims)
 {
+     matrixio_id data_id;
+     data_id.id = 0;
+     data_id.parallel = id.parallel;
 #if defined(HAVE_HDF5)
+ {
      int i, rank_copy;
-     hid_t space_id, data_id;
+     hid_t space_id;
      hsize_t *dims_copy, *maxdims;
 
-     data_id = H5Dopen(id, name);
+     CHECK((data_id.id = H5Dopen(id.id, name)) >= 0, "error in H5Dopen");
 
-     CHECK((data_id = H5Dopen(id, name)) >= 0, "error in H5Dopen");
-
-     CHECK((space_id = H5Dget_space(data_id)) >= 0,
+     CHECK((space_id = H5Dget_space(data_id.id)) >= 0,
 	   "error in H5Dget_space");
 
      rank_copy = H5Sget_simple_extent_ndims(space_id);
@@ -474,11 +519,9 @@ matrixio_id matrixio_open_dataset(matrixio_id id,
      free(dims_copy);
 
      H5Sclose(space_id);
-
-     return data_id;
-#else
-     return 0;
+ }
 #endif
+     return data_id;
 }
 
 /*****************************************************************************/
@@ -487,19 +530,27 @@ matrixio_id matrixio_create_dataset(matrixio_id id,
 				    const char *name, const char *description,
 				    int rank, const int *dims)
 {
+     matrixio_id data_id;
+     data_id.id = 0;
+     data_id.parallel = id.parallel;
 #if defined(HAVE_HDF5)
+ {
      int i;
-     hid_t space_id, type_id, data_id;
+     hid_t space_id, type_id;
      hsize_t *dims_copy;
 
      /* delete pre-existing datasets, or we'll have an error; I think
         we can only do this on the master process. (?) */
      if (matrixio_dataset_exists(id, name)) {
-	  if (mpi_is_master()) {
+#  ifdef HAVE_H5PSET_FAPL_MPIO /* H5Gunlink is collective */
+	  matrixio_dataset_delete(id, name);
+#  else
+	  if (mpi_is_master() || !id.parallel) {
 	       matrixio_dataset_delete(id, name);
-	       H5Fflush(id, H5F_SCOPE_GLOBAL);
+	       H5Fflush(id.id, H5F_SCOPE_GLOBAL);
 	  }
-	  IF_EXCLUSIVE(0,MPI_Barrier(MPI_COMM_WORLD));
+	  IF_EXCLUSIVE(0,if (id.parallel) MPI_Barrier(MPI_COMM_WORLD));
+#  endif
      }
 
      CHECK(rank > 0, "non-positive rank");
@@ -512,8 +563,10 @@ matrixio_id matrixio_create_dataset(matrixio_id id,
 
      free(dims_copy);
 
-#ifdef SCALAR_SINGLE_PREC
+#if defined(SCALAR_SINGLE_PREC)
      type_id = H5T_NATIVE_FLOAT;
+#elif defined(SCALAR_LONG_DOUBLE_PREC)
+     type_id = H5T_NATIVE_LDOUBLE;
 #else
      type_id = H5T_NATIVE_DOUBLE;
 #endif
@@ -521,26 +574,24 @@ matrixio_id matrixio_create_dataset(matrixio_id id,
      /* Create the dataset.  Note that, on parallel machines, H5Dcreate
 	should do the right thing; it is supposedly a collective operation. */
      IF_EXCLUSIVE(
-	  if (mpi_is_master())
-	       data_id = H5Dcreate(id, name, type_id, space_id, H5P_DEFAULT);
+	  if (mpi_is_master() || !id.parallel)
+	       data_id.id = H5Dcreate(id.id,name,type_id,space_id,H5P_DEFAULT);
 	  else
-	       data_id = H5Dopen(id, name),
-	  data_id = H5Dcreate(id, name, type_id, space_id, H5P_DEFAULT));
+	       data_id.id = H5Dopen(id.id, name),
+	  data_id.id = H5Dcreate(id.id, name, type_id, space_id, H5P_DEFAULT));
 
      H5Sclose(space_id);  /* the dataset should have its own copy now */
      
      matrixio_write_string_attr(data_id, "description", description);
-
-     return data_id;
-#else
-     return 0;
+ }
 #endif
+     return data_id;
 }
 
 void matrixio_close_dataset(matrixio_id data_id)
 {
 #if defined(HAVE_HDF5)
-     CHECK(H5Dclose(data_id) >= 0, "error closing HDF dataset");
+     CHECK(H5Dclose(data_id.id) >= 0, "error closing HDF dataset");
 #endif
 }
 
@@ -548,7 +599,7 @@ int matrixio_dataset_exists(matrixio_id id, const char *name)
 {
 #if defined(HAVE_HDF5)
      hid_t data_id;
-     SUPPRESS_HDF5_ERRORS(data_id = H5Dopen(id, name));
+     SUPPRESS_HDF5_ERRORS(data_id = H5Dopen(id.id, name));
      if (data_id >= 0)
 	  H5Dclose(data_id);
      return (data_id >= 0);
@@ -560,7 +611,7 @@ int matrixio_dataset_exists(matrixio_id id, const char *name)
 void matrixio_dataset_delete(matrixio_id id, const char *name)
 {
 #if defined(HAVE_HDF5)
-     H5Gunlink(id, name);
+     H5Gunlink(id.id, name);
 #endif
 }
 
@@ -575,7 +626,7 @@ void matrixio_write_real_data(matrixio_id data_id,
      int rank;
      hsize_t *dims, *maxdims;
      hid_t space_id, type_id, mem_space_id;
-     hssize_t *start;
+     start_t *start;
      hsize_t *strides, *count, count_prod;
      int i;
      real *data_copy;
@@ -584,7 +635,7 @@ void matrixio_write_real_data(matrixio_id data_id,
      /*******************************************************************/
      /* Get dimensions of dataset */
      
-     space_id = H5Dget_space(data_id);
+     space_id = H5Dget_space(data_id.id);
 
      rank = H5Sget_simple_extent_ndims(space_id);
      
@@ -595,8 +646,10 @@ void matrixio_write_real_data(matrixio_id data_id,
 
      free(maxdims);
 
-#ifdef SCALAR_SINGLE_PREC
+#if defined(SCALAR_SINGLE_PREC)
      type_id = H5T_NATIVE_FLOAT;
+#elif defined(SCALAR_LONG_DOUBLE_PREC)
+     type_id = H5T_NATIVE_LDOUBLE;
 #else
      type_id = H5T_NATIVE_DOUBLE;
 #endif
@@ -638,7 +691,7 @@ void matrixio_write_real_data(matrixio_id data_id,
      /* Before we can write the data to the data set, we must define
 	the dimensions and "selections" of the arrays to be read & written: */
 
-     CHK_MALLOC(start, hssize_t, rank);
+     CHK_MALLOC(start, start_t, rank);
      CHK_MALLOC(strides, hsize_t, rank);
      CHK_MALLOC(count, hsize_t, rank);
 
@@ -675,8 +728,8 @@ void matrixio_write_real_data(matrixio_id data_id,
      /* Write the data, then free all the stuff we've allocated. */
 
      if (do_write)
-	  H5Dwrite(data_id, type_id, mem_space_id, space_id, H5P_DEFAULT, 
-		   (void*) data_copy);
+	  H5Dwrite(data_id.id, type_id, mem_space_id, space_id, H5P_DEFAULT, 
+		   data_copy);
 
      if (free_data_copy)
 	  free(data_copy);
@@ -748,10 +801,10 @@ real *matrixio_read_real_data(matrixio_id id,
 	  strcpy(dname, name);
      }
      else {
-	  if (H5Giterate(id, "/", NULL, find_dataset, &dname) < 0)
+	  if (H5Giterate(id.id, "/", NULL, find_dataset, &dname) < 0)
 	       return NULL;
      }
-     SUPPRESS_HDF5_ERRORS(data_id = H5Dopen(id, dname));
+     SUPPRESS_HDF5_ERRORS(data_id = H5Dopen(id.id, dname));
      free(dname);
      if (data_id < 0)
 	  return NULL;
@@ -788,8 +841,10 @@ real *matrixio_read_real_data(matrixio_id id,
 	  for (i = 0; i < *rank; ++i)
 	       dims[i] = dims_copy[i];
 
-#ifdef SCALAR_SINGLE_PREC
+#if defined(SCALAR_SINGLE_PREC)
      type_id = H5T_NATIVE_FLOAT;
+#elif defined(SCALAR_LONG_DOUBLE_PREC)
+     type_id = H5T_NATIVE_LDOUBLE;
 #else
      type_id = H5T_NATIVE_DOUBLE;
 #endif
@@ -799,10 +854,10 @@ real *matrixio_read_real_data(matrixio_id id,
 	the dimensions and "selections" of the arrays to be read & written: */
 
      if (data) {
-	  hssize_t *start;
+	  start_t *start;
 	  hsize_t *strides, *count;
 
-	  CHK_MALLOC(start, hssize_t, *rank);
+	  CHK_MALLOC(start, start_t, *rank);
 	  CHK_MALLOC(strides, hsize_t, *rank);
 	  CHK_MALLOC(count, hsize_t, *rank);
 	  
@@ -845,7 +900,7 @@ real *matrixio_read_real_data(matrixio_id id,
      /* Read the data, then free all the H5 identifiers. */
 
      CHECK(H5Dread(data_id, type_id, mem_space_id, space_id, H5P_DEFAULT, 
-		   (void*) data) >= 0,
+		   data) >= 0,
 	   "error reading HDF5 dataset");
 
      if (mem_space_id != H5S_ALL)
