@@ -1,4 +1,4 @@
-/* Copyright (C) 1999, 2000, 2001, 2002, Massachusetts Institute of Technology.
+/* Copyright (C) 1999-2014 Massachusetts Institute of Technology.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -19,8 +19,9 @@
 #include <stdio.h>
 #include <math.h>
 
-#include "../config.h"
+#include "config.h"
 #include <check.h>
+#include <sphere-quad.h>
 #include <mpiglue.h>
 #include <mpi_utils.h>
 
@@ -36,8 +37,8 @@
 #  define HEEV F77_FUNC(zheev,ZHEEV)
 #  define SYEV F77_FUNC(dsyev,DSYEV)
 #endif
-extern void HEEV(char *, char *, int *, scalar *, int *, real *,
-		 scalar *, int *, real *, int *);
+extern void HEEV(char *, char *, int *, scalar_complex *, int *, real *,
+		 scalar_complex *, int *, real *, int *);
 extern void SYEV(char *, char *, int *, real *, int *, real *,
 		 real *, int *, int *);
 
@@ -46,7 +47,8 @@ void maxwell_sym_matrix_eigs(real eigs[3], const symmetric_matrix *V)
 {
      int n = 3, nw = 9, info;
 #if defined(WITH_HERMITIAN_EPSILON)
-     scalar_complex Vm[3][3], W[9], W2[9];
+     scalar_complex Vm[3][3], W[9];
+     real W2[9];
      CASSIGN_SCALAR(Vm[0][0], V->m00, 0);
      CASSIGN_SCALAR(Vm[1][1], V->m11, 0);
      CASSIGN_SCALAR(Vm[2][2], V->m22, 0);
@@ -152,7 +154,7 @@ void maxwell_sym_matrix_invert(symmetric_matrix *Vinv,
 }
 
 /* Returns whether or not V is positive-definite. */
-static int sym_matrix_positive_definite(symmetric_matrix *V)
+int maxwell_sym_matrix_positive_definite(symmetric_matrix *V)
 {
      real det2, det3;
      real m00 = V->m00, m11 = V->m11, m22 = V->m22;
@@ -174,6 +176,60 @@ static int sym_matrix_positive_definite(symmetric_matrix *V)
      return (m00 > 0.0 && det2 > 0.0 && det3 > 0.0);
 }
 
+#define EQ(x1,x2) (fabs((x1) - (x2)) < tol)
+static int sym_matrix_eq(symmetric_matrix V1, symmetric_matrix V2, double tol)
+{
+     if (!EQ(V1.m00,V2.m00) || !EQ(V1.m11,V2.m11) || !EQ(V1.m22,V2.m22))
+	  return 0;
+#if defined(WITH_HERMITIAN_EPSILON)
+     return(EQ(V1.m01.re,V2.m01.re) && EQ(V1.m01.im,V2.m01.im) &&
+	    EQ(V1.m02.re,V2.m02.re) && EQ(V1.m02.im,V2.m02.im) &&
+	    EQ(V1.m12.re,V2.m12.re) && EQ(V1.m12.im,V2.m12.im));
+#else
+     return(EQ(V1.m01,V2.m01) && EQ(V1.m02,V2.m02) && EQ(V1.m12,V2.m12));
+#endif
+}
+
+/* rotate A by a unitary (real) rotation matrix R:
+      RAR = transpose(R) * A * R
+*/
+void maxwell_sym_matrix_rotate(symmetric_matrix *RAR,
+			       const symmetric_matrix *A_,
+			       const double R[3][3])
+{
+     int i,j;
+     double A[3][3], AR[3][3];
+     A[0][0] = A_->m00;
+     A[1][1] = A_->m11;
+     A[2][2] = A_->m22;
+     A[0][1] = A[1][0] = ESCALAR_RE(A_->m01);
+     A[0][2] = A[2][0] = ESCALAR_RE(A_->m02);
+     A[1][2] = A[2][1] = ESCALAR_RE(A_->m12);
+     for (i = 0; i < 3; ++i) for (j = 0; j < 3; ++j) 
+	  AR[i][j] = A[i][0]*R[0][j] + A[i][1]*R[1][j] + A[i][2]*R[2][j];
+     for (i = 0; i < 3; ++i) for (j = i; j < 3; ++j) 
+	  A[i][j] = R[0][i]*AR[0][j] + R[1][i]*AR[1][j] + R[2][i]*AR[2][j];
+     RAR->m00 = A[0][0];
+     RAR->m11 = A[1][1];
+     RAR->m22 = A[2][2];
+     ESCALAR_RE(RAR->m01) = A[0][1];
+     ESCALAR_RE(RAR->m02) = A[0][2];
+     ESCALAR_RE(RAR->m12) = A[1][2];
+#if defined(WITH_HERMITIAN_EPSILON)
+     A[0][0] = A[1][1] = A[2][2] = 0;
+     A[1][0] = -(A[0][1] = ESCALAR_IM(A_->m01));
+     A[2][0] = -(A[0][2] = ESCALAR_IM(A_->m02));
+     A[2][1] = -(A[1][2] = ESCALAR_IM(A_->m12));
+     for (i = 0; i < 3; ++i) for (j = 0; j < 3; ++j) 
+	  AR[i][j] = A[i][0]*R[0][j] + A[i][1]*R[1][j] + A[i][2]*R[2][j];
+     for (i = 0; i < 3; ++i) for (j = i; j < 3; ++j) 
+	  A[i][j] = R[0][i]*AR[0][j] + R[1][i]*AR[1][j] + R[2][i]*AR[2][j];
+     ESCALAR_IM(RAR->m01) = A[0][1];
+     ESCALAR_IM(RAR->m02) = A[0][2];
+     ESCALAR_IM(RAR->m12) = A[1][2];
+#endif
+}
+
 /**************************************************************************/
 
 int check_maxwell_dielectric(maxwell_data *d,
@@ -185,7 +241,7 @@ int check_maxwell_dielectric(maxwell_data *d,
 
      for (i = 0; i < d->fft_output_size; ++i) {
 	  if (!negative_epsilon_okp &&
-	      !sym_matrix_positive_definite(d->eps_inv + i))
+	      !maxwell_sym_matrix_positive_definite(d->eps_inv + i))
 	       return 1;
 	  if (require_2d) {
 #if defined(WITH_HERMITIAN_EPSILON)
@@ -205,155 +261,12 @@ int check_maxwell_dielectric(maxwell_data *d,
 
 /**************************************************************************/
 
-#define SHIFT3(x,y,z) {real SHIFT3_dummy = z; z = y; y = x; x = SHIFT3_dummy;}
-
-/* Compute quadrature points and weights for integrating on the
-   unit sphere.  x, y, z, and weight should be arrays of num_sq_pts
-   values to hold the coordinates and weights of the quadrature points
-   on output.   Currently, num_sq_pts = 12, 50, and 72 are supported. */
-void spherical_quadrature_points(real *x, real *y, real *z,
-				 real *weight, int num_sq_pts)
-{
-     int i,j,k,l, n = 0;
-     real x0, y0, z0, w;
-
-     if (num_sq_pts == 50) {
-	  /* Computes quadrature points and weights for 50-point 11th degree
-	     integration formula on a unit sphere.  This particular quadrature
-	     formula has the advantage, for our purposes, of preserving the
-	     symmetry group of an octahedraon (i.e. simple cubic symmetry, with
-	     respect to the Cartesian xyz axes).  References:
-	
-	     A. D. McLaren, "Optimal Numerical Integration on a Sphere,"
-	     Math. Comp. 17, pp. 361-383 (1963).
-	     
-	     Also in: Arthur H. Stroud, "Approximate Calculation of Multiple
-	     Integrals" (Prentice Hall, 1971) (formula number U3:11-1). 
-
-	     This code was written with the help of example code by
-	     John Burkardt: 
-	           http://www.psc.edu/~burkardt/src/stroud/stroud.html */
-     
-	  x0 = 1; y0 = z0 = 0;
-	  w = 9216 / 725760.0;
-	  for (i = 0; i < 2; ++i) {
-	       x0 = -x0;
-	       for (j = 0; j < 3; ++j) {
-		    SHIFT3(x0,y0,z0);
-		    x[n] = x0; y[n] = y0; z[n] = z0; weight[n++] = w;
-	       }
-	  }
-	  
-	  x0 = y0 = sqrt(0.5); z0 = 0;
-	  w = 16384 / 725760.0;
-	  for (i = 0; i < 2; ++i) {
-	       x0 = -x0;
-	       for (j = 0; j < 2; ++j) {
-		    y0 = -y0;
-		    for (k = 0; k < 3; ++k) {
-			 SHIFT3(x0,y0,z0);
-			 x[n] = x0; y[n] = y0; z[n] = z0; weight[n++] = w;
-		    }
-	       }
-	  }
-	  
-	  x0 = y0 = z0 = sqrt(1.0 / 3.0);
-	  w = 15309 / 725760.0;
-	  for (i = 0; i < 2; ++i) {
-	       x0 = -x0;
-	       for (j = 0; j < 2; ++j) {
-		    y0 = -y0;
-		    for (k = 0; k < 2; ++k) {
-			 z0 = -z0;
-			 x[n] = x0; y[n] = y0; z[n] = z0; weight[n++] = w;
-		    }
-	       }
-	  }
-	  
-	  x0 = y0 = sqrt(1.0 / 11.0); z0 = 3 * x0;
-	  w = 14641 / 725760.0;
-	  for (i = 0; i < 2; ++i) {
-	       x0 = -x0;
-	       for (j = 0; j < 2; ++j) {
-		    y0 = -y0;
-		    for (k = 0; k < 2; ++k) {
-			 z0 = -z0;
-			 for (l = 0; l < 3; ++l) {
-			      SHIFT3(x0,y0,z0);
-			      x[n] = x0; y[n] = y0; z[n] = z0; weight[n++] = w;
-			 }
-		    }
-	       }
-	  }
-     }  
-     else if (num_sq_pts == 72 || num_sq_pts == 12) {
-	  /* As above (same references), but with a 72-point 14th degree
-	     formula, this time with the symmetry group of an icosohedron.
-	     (Stroud formula number U3:14-1.)  Alternatively, just use
-	     the 12-point 5th degree formula consisting of the vertices
-	     of a regular icosohedron. */
-	  
-	  /* first, the vertices of an icosohedron: */
-	  x0 = sqrt(0.5 - sqrt(0.05));
-	  y0 = sqrt(0.5 + sqrt(0.05));
-	  z0 = 0;
-	  if (num_sq_pts == 72)
-	       w = 125 / 10080.0;
-	  else
-	       w = 1 / 12.0;
-	  for (i = 0; i < 2; ++i) {
-	       x0 = -x0;
-	       for (j = 0; j < 2; ++j) {
-		    y0 = -y0;
-		    for (k = 0; k < 3; ++k) {
-			 SHIFT3(x0,y0,z0);
-			 x[n] = x0; y[n] = y0; z[n] = z0; weight[n++] = w;
-		    }
-	       }
-	  }
-	  
-	  if (num_sq_pts == 72) {
-	       /* it would be nice, for completeness, to have more
-                  digits here: */
-	       real coords[3][5] = {
-		    { -0.151108275, 0.315838353, 0.346307112, -0.101808787, -0.409228403 },
-		    { 0.155240600, 0.257049387, 0.666277790,  0.817386065, 0.501547712 },
-		    { 0.976251323, 0.913330032, 0.660412970,  0.567022920, 0.762221757 }
-	       };
-	       
-	       w = 143 / 10080.0;
-	       for (l = 0; l < 5; ++l) {
-		    x0 = coords[0][l]; y0 = coords[1][l]; z0 = coords[2][l];
-		    for (i = 0; i < 3; ++i) {
-			 double dummy = x0;
-			 x0 = z0;
-			 z0 = -y0;
-			 y0 = -dummy;
-			 for (j = 0; j < 3; ++j) {
-			      SHIFT3(x0,y0,z0);
-			      x[n] = x0; y[n] = y0; z[n] = z0; weight[n++] = w;
-			 }
-			 y0 = -y0;
-			 z0 = -z0;
-			 x[n] = x0; y[n] = y0; z[n] = z0; weight[n++] = w;
-		    }
-	       }
-	  }
-     }
-     else
-	  mpi_die("spherical_quadrature_points: passed unknown # points!");
-
-     CHECK(n == num_sq_pts,
-	   "bug in spherical_quadrature_points: wrong number of points!");
-}
-
 #define K_PI 3.141592653589793238462643383279502884197
 #define SMALL 1.0e-6
 #define MAX2(a,b) ((a) > (b) ? (a) : (b))
 #define MIN2(a,b) ((a) < (b) ? (a) : (b))
 
-#define NUM_SQ_PTS 50  /* number of spherical quadrature points to use */
-#define MAX_MOMENT_MESH NUM_SQ_PTS /* max # of moment-mesh vectors */
+#define MAX_MOMENT_MESH NQUAD /* max # of moment-mesh vectors */
 #define MOMENT_MESH_R 0.5
 
 /* A function to set up the mesh given the grid dimensions, mesh size,
@@ -391,44 +304,11 @@ static void get_mesh(int nx, int ny, int nz, const int mesh_size[3],
 	  *mesh_prod *= ms;
      }
 
-     /* Set the mesh to compute the "dipole moment" of epsilon over,
-	in Cartesian coordinates (used to compute the normal vector
-	at surfaces).  Ideally, a spherically-symmetrical distribution
-	of points on the radius MOMENT_MESH_R sphere. */
-     if (rank == 1) {
-	  /* one-dimensional: just two points (really, normal
-	     vector is always along x): */
-	  *size_moment_mesh = 2;
-	  moment_mesh[0][0] = -MOMENT_MESH_R;
-	  moment_mesh[0][1] = 0.0;
-	  moment_mesh[0][2] = 0.0;
-	  moment_mesh[1][0] = MOMENT_MESH_R;
-	  moment_mesh[1][1] = 0.0;
-	  moment_mesh[1][2] = 0.0;
-	  moment_mesh_weights[0] = moment_mesh_weights[1] = 0.5;
-     }
-     else if (rank == 2) {
-	  /* two-dimensional: 12 points at 30-degree intervals: */
-	  *size_moment_mesh = 12;
-	  for (i = 0; i < *size_moment_mesh; ++i) {
-	       moment_mesh[i][0] =
-		    cos(2*i * K_PI / *size_moment_mesh) * MOMENT_MESH_R;
-	       moment_mesh[i][1] =
-		    sin(2*i * K_PI / *size_moment_mesh) * MOMENT_MESH_R;
-	       moment_mesh[i][2] = 0.0;
-	       moment_mesh_weights[i] = 1.0 / *size_moment_mesh;
-	  }
-     }
-     else {
-	  real x[NUM_SQ_PTS], y[NUM_SQ_PTS], z[NUM_SQ_PTS];
-
-	  *size_moment_mesh = NUM_SQ_PTS;
-	  spherical_quadrature_points(x,y,z, moment_mesh_weights, NUM_SQ_PTS);
-	  for (i = 0; i < *size_moment_mesh; ++i) {
-	       moment_mesh[i][0] = x[i] * MOMENT_MESH_R;
-	       moment_mesh[i][1] = y[i] * MOMENT_MESH_R;
-	       moment_mesh[i][2] = z[i] * MOMENT_MESH_R;
-	  }
+     *size_moment_mesh = num_sphere_quad[rank-1];
+     for (i = 0; i < num_sphere_quad[rank-1]; ++i) {
+	  for (j = 0; j < 3; ++j)
+	       moment_mesh[i][j] = sphere_quad[rank-1][i][j] * MOMENT_MESH_R;
+	  moment_mesh_weights[i] = sphere_quad[rank-1][i][3];
      }
 
      CHECK(*size_moment_mesh <= MAX_MOMENT_MESH, "yikes, moment mesh too big");
@@ -503,6 +383,7 @@ void set_maxwell_dielectric(maxwell_data *md,
 			    const int mesh_size[3],
 			    real R[3][3], real G[3][3],
 			    maxwell_dielectric_function epsilon,
+			    maxwell_dielectric_mean_function mepsilon,
 			    void *epsilon_data)
 {
      real s1, s2, s3, m1, m2, m3;  /* grid/mesh steps */
@@ -529,20 +410,12 @@ void set_maxwell_dielectric(maxwell_data *md,
 	      &size_moment_mesh);
      mesh_prod_inv = 1.0 / mesh_prod;
 
-#ifdef DEBUG
-     mpi_one_printf("Using moment mesh (%d):\n", size_moment_mesh);
-     for (i = 0; i < size_moment_mesh; ++i)
-	  mpi_one_printf("   (%g, %g, %g) (%g)\n",
-		 moment_mesh[i][0], moment_mesh[i][1], moment_mesh[i][2],
-		 moment_mesh_weights[i]);
-#endif
-
      s1 = 1.0 / n1;
      s2 = 1.0 / n2;
      s3 = 1.0 / n3;
-     m1 = s1 / MAX2(1, mesh_size[0] - 1);
-     m2 = s2 / MAX2(1, mesh_size[1] - 1);
-     m3 = s3 / MAX2(1, mesh_size[2] - 1);
+     m1 = s1 / MAX2(1, mesh_size[0]);
+     m2 = s2 / MAX2(1, mesh_size[1]);
+     m3 = s3 / MAX2(1, mesh_size[2]);
 
      /* Here we have different loops over the coordinates, depending
 	upon whether we are using complex or real and serial or
@@ -585,6 +458,8 @@ void set_maxwell_dielectric(maxwell_data *md,
 
 #  ifndef HAVE_MPI
 
+     (void) k; /* unused */
+
      n_other = md->other_dims;
      n_last = md->last_dim_size / 2;
      rank = (n3 == 1) ? (n2 == 1 ? 1 : 2) : 3;
@@ -630,17 +505,74 @@ void set_maxwell_dielectric(maxwell_data *md,
      {
 	  int mi, mj, mk;
 #ifdef WITH_HERMITIAN_EPSILON
-	  symmetric_matrix eps_mean = {0,0,0,{0,0},{0,0},{0,0}}, 
-			   eps_inv_mean = {0,0,0,{0,0},{0,0},{0,0}},
-			   eps_mean_inv;
+	  symmetric_matrix eps_mean, eps_inv_mean, eps_mean_inv;
 #else
-	  symmetric_matrix eps_mean = {0,0,0,0,0,0}, 
-			   eps_inv_mean = {0,0,0,0,0,0}, eps_mean_inv;
+	  symmetric_matrix eps_mean, eps_inv_mean, eps_mean_inv;
 #endif
 	  real norm_len;
 	  real norm0, norm1, norm2;
 	  short means_different_p, diag_eps_p;
-	  
+
+	  {
+	       real r[3], normal[3];
+	       r[0] = i2 * s1;
+	       r[1] = j2 * s2;
+	       r[2] = k2 * s3;
+	       if (mepsilon && mepsilon(&eps_mean, &eps_inv_mean, normal,
+					s1, s2, s3, mesh_prod_inv,
+					r, epsilon_data)) {
+
+#ifdef KOTTKE /* mepsilon did new anisotropic smoothing w/Kottke algorithm */
+		    maxwell_sym_matrix_invert(md->eps_inv + eps_index, 
+					      &eps_mean);
+		    goto got_eps_inv;
+#endif
+
+		    norm0 = R[0][0] * normal[0] + R[1][0] * normal[1]
+			 + R[2][0] * normal[2];
+		    norm1 = R[0][1] * normal[0] + R[1][1] * normal[1]
+			 + R[2][1] * normal[2];
+		    norm2 = R[0][2] * normal[0] + R[1][2] * normal[1]
+			 + R[2][2] * normal[2];
+		    means_different_p = 1;
+		    diag_eps_p = DIAG_SYMMETRIC_MATRIX(eps_mean);
+		    maxwell_sym_matrix_invert(&eps_mean_inv, &eps_mean);
+
+#if !defined(SCALAR_COMPLEX) && 0 /* check inversion symmetry */
+		    {
+			 symmetric_matrix eps_mean2, eps_inv_mean2;
+			 real normal2[3], r2[3], nc[3];
+			 r2[0] = n1 == 0 ? r[0] : 1.0 - r[0];
+			 r2[1] = n2 == 0 ? r[1] : 1.0 - r[1];
+			 r2[2] = n3 == 0 ? r[2] : 1.0 - r[2];
+			 CHECK(mepsilon(&eps_mean2, &eps_inv_mean2, normal2,
+					s1, s2, s3, mesh_prod_inv,
+					r2, epsilon_data),
+			       "mepsilon symmetry is broken");
+			 CHECK(sym_matrix_eq(eps_mean,eps_mean2,1e-10) &&
+			       sym_matrix_eq(eps_inv_mean,eps_inv_mean2,1e-10),
+			       "inversion symmetry is broken");
+			 nc[0] = normal[1]*normal2[2] - normal[2]*normal2[1];
+			 nc[1] = normal[2]*normal2[0] - normal[0]*normal2[2];
+			 nc[2] = normal[0]*normal2[1] - normal[1]*normal2[0];
+			 CHECK(sqrt(nc[0]*nc[0]+nc[1]*nc[1]+nc[2]*nc[2])<1e-6,
+			       "normal-vector symmetry is broken");
+		    }
+#endif
+
+		    goto got_mean;
+	       }
+	  }
+
+	  eps_mean.m00 = eps_mean.m11 = eps_mean.m22 =
+	       eps_inv_mean.m00 = eps_inv_mean.m11 = eps_inv_mean.m22 = 0.0;
+	  ASSIGN_ESCALAR(eps_mean.m01, 0,0);
+	  ASSIGN_ESCALAR(eps_mean.m02, 0,0);
+	  ASSIGN_ESCALAR(eps_mean.m12, 0,0);
+	  ASSIGN_ESCALAR(eps_inv_mean.m01, 0,0);
+	  ASSIGN_ESCALAR(eps_inv_mean.m02, 0,0);
+	  ASSIGN_ESCALAR(eps_inv_mean.m12, 0,0);
+
 	  for (mi = 0; mi < mesh_size[0]; ++mi)
 	       for (mj = 0; mj < mesh_size[1]; ++mj)
 		    for (mk = 0; mk < mesh_size[2]; ++mk) {
@@ -763,11 +695,14 @@ void set_maxwell_dielectric(maxwell_data *md,
 		    moment1 += eps_trace * moment_mesh[mi][1];
 		    moment2 += eps_trace * moment_mesh[mi][2];
 	       }
+
 	       /* need to convert moment from lattice to cartesian coords: */
 	       norm0 = R[0][0]*moment0 + R[1][0]*moment1 + R[2][0]*moment2;
 	       norm1 = R[0][1]*moment0 + R[1][1]*moment1 + R[2][1]*moment2;
 	       norm2 = R[0][2]*moment0 + R[1][2]*moment1 + R[2][2]*moment2;
 	  
+	  got_mean:
+
 	       norm_len = sqrt(norm0*norm0 + norm1*norm1 + norm2*norm2);
 	  }
 	  
@@ -866,6 +801,7 @@ void set_maxwell_dielectric(maxwell_data *md,
 	  else { /* undetermined normal vector and/or constant eps */
 	       md->eps_inv[eps_index] = eps_mean_inv;
 	  }
+     got_eps_inv:
 	  
 	  eps_inv_total += (md->eps_inv[eps_index].m00 + 
 			    md->eps_inv[eps_index].m11 + 
