@@ -1,4 +1,4 @@
-/* Copyright (C) 1999, 2000, 2001, 2002, Massachusetts Institute of Technology.
+/* Copyright (C) 1999-2014 Massachusetts Institute of Technology.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -19,11 +19,9 @@
 #include <stdio.h>
 #include <math.h>
 
-#include "../config.h"
+#include "imaxwell.h"
 #include <check.h>
 
-#include "maxwell.h"
-
 /**************************************************************************/
 
 /* assign a = v going from transverse to cartesian coordinates.  
@@ -155,28 +153,143 @@ static void assign_ucross_t2c(scalar *a, const real u[3], const k_data k,
 
 /**************************************************************************/
 
-void maxwell_compute_fft(int dir, maxwell_data *d, scalar *array, 
+void maxwell_compute_fft(int dir, maxwell_data *d, 
+			 scalar *array_in, scalar *array_out, 
 			 int howmany, int stride, int dist)
 {
-#ifdef HAVE_FFTW
+#if defined(HAVE_FFTW3)
+     FFTW(plan) plan, iplan;
+     FFTW(complex) *carray_in = (FFTW(complex) *) array_in;
+     real *rarray_in = (real *) array_in;
+     FFTW(complex) *carray_out = (FFTW(complex) *) array_out;
+     real *rarray_out = (real *) array_out;
+     int ip;
+     for (ip = 0; ip < d->nplans && (howmany != d->plans_howmany[ip] ||
+				     stride != d->plans_stride[ip] ||
+				     dist != d->plans_dist[ip]); ++ip);
+     if (ip < d->nplans) {
+	  plan = (FFTW(plan)) d->plans[ip];
+	  iplan = (FFTW(plan)) d->iplans[ip];
+     }
+     else { /* create new plans */
+	  ptrdiff_t np[3];
+	  int n[3]; np[0]=n[0]=d->nx; np[1]=n[1]=d->ny; np[2]=n[2]=d->nz;
+#  ifdef SCALAR_COMPLEX
+#    ifdef HAVE_MPI
+	  CHECK(stride==howmany && dist==1, "bug: unsupported stride/dist");
+	  plan = FFTW(mpi_plan_many_dft)(3, np, howmany, 
+					 FFTW_MPI_DEFAULT_BLOCK,
+					 FFTW_MPI_DEFAULT_BLOCK,
+					 carray_in, carray_out,
+					 MPI_COMM_WORLD, FFTW_BACKWARD,
+					 FFTW_ESTIMATE
+					 | FFTW_MPI_TRANSPOSED_IN);
+	  iplan = FFTW(mpi_plan_many_dft)(3, np, howmany, 
+					  FFTW_MPI_DEFAULT_BLOCK,
+					  FFTW_MPI_DEFAULT_BLOCK,
+					  carray_in, carray_out,
+					  MPI_COMM_WORLD, FFTW_FORWARD,
+					  FFTW_ESTIMATE
+					  | FFTW_MPI_TRANSPOSED_OUT);
+#    else /* !HAVE_MPI */
+	  plan = FFTW(plan_many_dft)(3, n, howmany, carray_in, 0, stride, dist,
+				     carray_out, 0, stride, dist,
+				     FFTW_BACKWARD, FFTW_ESTIMATE);
+	  iplan = FFTW(plan_many_dft)(3, n, howmany, carray_in,0,stride, dist,
+				      carray_out, 0, stride, dist,
+				      FFTW_FORWARD, FFTW_ESTIMATE);
+#    endif /* !HAVE_MPI */
+#  else /* !SCALAR_COMPLEX */
+	  {
+	       int rnk = n[2] != 1 ? 3 : (n[1] != 1 ? 2 : 1);
+	       int nr[3]; nr[0] = n[0]; nr[1] = n[1]; nr[2] = n[2];
+	       nr[rnk-1] = 2*(nr[rnk-1]/2 + 1);
+#    ifdef HAVE_MPI
+	  CHECK(stride==howmany && dist==1, "bug: unsupported stride/dist");
+	  plan = FFTW(mpi_plan_many_dft_c2r)(rnk, np, howmany, 
+					     FFTW_MPI_DEFAULT_BLOCK,
+					     FFTW_MPI_DEFAULT_BLOCK,
+					     carray_in, rarray_out,
+					     MPI_COMM_WORLD, FFTW_ESTIMATE
+					     | FFTW_MPI_TRANSPOSED_IN);
+	  iplan = FFTW(mpi_plan_many_dft_r2c)(rnk, np, howmany, 
+					      FFTW_MPI_DEFAULT_BLOCK,
+					      FFTW_MPI_DEFAULT_BLOCK,
+					      rarray_in, carray_out,
+					      MPI_COMM_WORLD, FFTW_ESTIMATE
+					      | FFTW_MPI_TRANSPOSED_OUT);
+#    else /* !HAVE_MPI */
+	       plan = FFTW(plan_many_dft_c2r)(rnk, n, howmany,
+					      carray_in, 0, stride, dist,
+					      rarray_out, nr, stride, dist,
+					      FFTW_ESTIMATE);
+	       iplan = FFTW(plan_many_dft_r2c)(rnk, n, howmany,
+					       rarray_in, nr, stride, dist,
+					       carray_out, 0, stride, dist,
+					       FFTW_ESTIMATE);
+#    endif /* !HAVE_MPI */
+	  }
+#  endif /* !SCALAR_COMPLEX */
+	  CHECK(plan && iplan, "Failure creating FFTW3 plans");
+     }
+
+     /* note that the new-array execute functions should be safe
+	since we only apply maxwell_compute_fft to fftw_malloc'ed data 
+	(so we don't ever have misaligned arrays), and we check above
+	that the strides etc. match */
+#  ifdef SCALAR_COMPLEX
+#    ifdef HAVE_MPI
+     FFTW(mpi_execute_dft)(dir < 0 ? plan : iplan, carray_in, carray_out);
+#    else /* !HAVE_MPI */
+     FFTW(execute_dft)(dir < 0 ? plan : iplan, carray_in, carray_out);
+#    endif /* !HAVE_MPI */
+#  else
+#    ifdef HAVE_MPI
+     if (dir > 0)
+	  FFTW(mpi_execute_dft_r2c)(iplan, rarray_in, carray_out);
+     else
+	  FFTW(mpi_execute_dft_c2r)(plan, carray_in, rarray_out);
+#    else /* !HAVE_MPI */
+     if (dir > 0)
+	  FFTW(execute_dft_r2c)(iplan, rarray_in, carray_out);
+     else
+	  FFTW(execute_dft_c2r)(plan, carray_in, rarray_out);
+#    endif /* !HAVE_MPI */
+#  endif
+
+     if (ip == MAX_NPLANS) { /* don't store too many plans */
+	  FFTW(destroy_plan)(plan);
+	  FFTW(destroy_plan)(iplan);
+     }
+     else if (ip == d->nplans) { /* save for later re-use */
+	  d->plans[ip] = plan;
+	  d->iplans[ip] = iplan;
+	  d->plans_howmany[ip] = howmany;
+	  d->plans_stride[ip] = stride;
+	  d->plans_dist[ip] = dist;
+	  d->nplans++;
+     }
+#elif defined(HAVE_FFTW)
+
+     CHECK(array_in == array_out, "only in-place supported with FFTW2");
 
 #  ifdef SCALAR_COMPLEX
 
 #    ifndef HAVE_MPI
 
-     fftwnd(dir < 0 ? d->plan : d->iplan,
+     fftwnd((fftplan) (dir < 0 ? d->plans[0] : d->iplans[0]),
 	    howmany,
-	    (fftw_complex *) array, stride, dist,
+	    (fftw_complex *) array_in, stride, dist,
 	    0, 0, 0);
 
 #    else /* HAVE_MPI */
 
      CHECK(stride == howmany && dist == 1,
 	   "weird strides and dists don't work with fftwnd_mpi");
-
-     fftwnd_mpi(dir < 0 ? d->plan : d->iplan,
+     
+     fftwnd_mpi((fftplan) (dir < 0 ? d->plans[0] : d->iplans[0]),
 		howmany,
-		(fftw_complex *) array, (fftw_complex *) NULL,
+		(fftw_complex *) array_in, (fftw_complex *) NULL,
 		FFTW_TRANSPOSED_ORDER);
 
 #    endif /* HAVE_MPI */
@@ -186,14 +299,14 @@ void maxwell_compute_fft(int dir, maxwell_data *d, scalar *array,
 #    ifndef HAVE_MPI
 
      if (dir > 0)
-	  rfftwnd_real_to_complex(d->iplan,
+	  rfftwnd_real_to_complex((fftplan) (d->iplans[0]),
 				  howmany,
-				  (fftw_real *) array, stride, dist,
+				  (fftw_real *) array_in, stride, dist,
 				  0, 0, 0);
      else
-	  rfftwnd_complex_to_real(d->plan,
+	  rfftwnd_complex_to_real((fftplan) (d->plans[0]),
 				  howmany,
-				  (fftw_complex *) array, stride, dist,
+				  (fftw_complex *) array_in, stride, dist,
 				  0, 0, 0);
 
 #    else /* HAVE_MPI */
@@ -201,8 +314,8 @@ void maxwell_compute_fft(int dir, maxwell_data *d, scalar *array,
      CHECK(stride == howmany && dist == 1,
 	   "weird strides and dists don't work with rfftwnd_mpi");
      
-     rfftwnd_mpi(dir < 0 ? d->plan : d->iplan,
-		 howmany, array, (scalar *) NULL,
+     rfftwnd_mpi((fftplan) (dir < 0 ? d->plans[0] : d->iplans[0]),
+		 howmany, array_in, (scalar *) NULL,
 		 FFTW_TRANSPOSED_ORDER);
 
 #    endif /* HAVE_MPI */
@@ -275,6 +388,7 @@ void maxwell_compute_d_from_H(maxwell_data *d, evectmatrix Hin,
 			      int cur_band_start, int cur_num_bands)
 {
      scalar *fft_data = (scalar *) dfield;
+     scalar *fft_data_in = d->fft_data2 == d->fft_data ? fft_data : (fft_data == d->fft_data ? d->fft_data2 : d->fft_data);
      int i, j, b;
 
      CHECK(Hin.c == 2, "fields don't have 2 components!");
@@ -291,7 +405,7 @@ void maxwell_compute_d_from_H(maxwell_data *d, evectmatrix Hin,
 	       k_data cur_k = d->k_plus_G[ij];
 	       
 	       for (b = 0; b < cur_num_bands; ++b)
-		    assign_cross_t2c(&fft_data[3 * (ij2*cur_num_bands 
+		    assign_cross_t2c(&fft_data_in[3 * (ij2*cur_num_bands 
 						    + b)], 
 				     cur_k, 
 				     &Hin.data[ij * 2 * Hin.p + 
@@ -300,7 +414,7 @@ void maxwell_compute_d_from_H(maxwell_data *d, evectmatrix Hin,
 	  }
 
      /* now, convert to position space via FFT: */
-     maxwell_compute_fft(+1, d, fft_data,
+     maxwell_compute_fft(+1, d, fft_data_in, fft_data,
 			 cur_num_bands*3, cur_num_bands*3, 1);
 }
 
@@ -339,6 +453,7 @@ void maxwell_compute_H_from_e(maxwell_data *d, evectmatrix Hout,
 			      real scale)
 {
      scalar *fft_data = (scalar *) efield;
+     scalar *fft_data_out = d->fft_data2 == d->fft_data ? fft_data : (fft_data == d->fft_data ? d->fft_data2 : d->fft_data);
      int i, j, b;
 
      CHECK(Hout.c == 2, "fields don't have 2 components!");
@@ -348,7 +463,7 @@ void maxwell_compute_H_from_e(maxwell_data *d, evectmatrix Hout,
 	   "invalid range of bands for computing fields");
 
      /* convert back to Fourier space */
-     maxwell_compute_fft(-1, d, fft_data,
+     maxwell_compute_fft(-1, d, fft_data, fft_data_out,
 			 cur_num_bands*3, cur_num_bands*3, 1);
      
      /* then, compute Hout = curl(fft_data) (* scale factor): */
@@ -363,7 +478,7 @@ void maxwell_compute_H_from_e(maxwell_data *d, evectmatrix Hout,
 		    assign_cross_c2t(&Hout.data[ij * 2 * Hout.p + 
 					       b + cur_band_start],
 				     Hout.p, cur_k, 
-				     &fft_data[3 * (ij2*cur_num_bands + b)],
+				     &fft_data_out[3 * (ij2*cur_num_bands+b)],
 				     scale);
 	  }
 }
@@ -376,6 +491,7 @@ void maxwell_compute_h_from_H(maxwell_data *d, evectmatrix Hin,
 			      int cur_band_start, int cur_num_bands)
 {
      scalar *fft_data = (scalar *) hfield;
+     scalar *fft_data_in = d->fft_data2 == d->fft_data ? fft_data : (fft_data == d->fft_data ? d->fft_data2 : d->fft_data);
      int i, j, b;
 
      CHECK(Hin.c == 2, "fields don't have 2 components!");
@@ -393,7 +509,7 @@ void maxwell_compute_h_from_H(maxwell_data *d, evectmatrix Hin,
                k_data cur_k = d->k_plus_G[ij];
 	       
 	       for (b = 0; b < cur_num_bands; ++b)
-		    assign_t2c(&fft_data[3 * (ij2*cur_num_bands 
+		    assign_t2c(&fft_data_in[3 * (ij2*cur_num_bands 
 					      + b)], 
 			       cur_k,
 			       &Hin.data[ij * 2 * Hin.p + 
@@ -402,7 +518,7 @@ void maxwell_compute_h_from_H(maxwell_data *d, evectmatrix Hin,
 	  }
 
      /* now, convert to position space via FFT: */
-     maxwell_compute_fft(+1, d, fft_data,
+     maxwell_compute_fft(+1, d, fft_data_in, fft_data,
 			 cur_num_bands*3, cur_num_bands*3, 1);
 }
 
@@ -660,6 +776,205 @@ void maxwell_vectorfield_otherhalf(maxwell_data *d, scalar_complex *field,
 #endif /* ! SCALAR_COMPLEX */
 }
 
+/* as vectorfield_otherhalf, but operates on a complex scalar field 
+   ... ugh, copy & paste job */
+void maxwell_cscalarfield_otherhalf(maxwell_data *d, scalar_complex *field,
+				    real phasex, real phasey, real phasez)
+{
+#ifndef SCALAR_COMPLEX
+     int i, j, jmin = 1;
+     int rank, n_other, n_last, n_last_stored, n_last_new, nx, ny, nz, nxmax;
+#  ifdef HAVE_MPI
+     int local_x_start;
+#  endif
+     scalar_complex pz, pxz, pyz, pxyz;
+
+     nxmax = nx = d->nx; ny = d->ny; nz = d->nz;
+     n_other = d->other_dims;
+     n_last = d->last_dim;
+     n_last_stored = d->last_dim_size / 2;
+     n_last_new = n_last - n_last_stored; /* < n_last_stored always */
+     rank = (nz == 1) ? (ny == 1 ? 1 : 2) : 3;
+
+#  ifdef HAVE_MPI
+     local_x_start = d->local_y_start;
+     CHECK(rank == 2 || rank == 3, "unsupported rfftwnd_mpi rank!");
+     if (rank == 2) {
+	  n_other = nx;
+	  n_last_new = ny = d->local_ny;
+	  if (local_x_start == 0)
+	       --n_last_new; /* DC frequency should not be in other half */
+	  else
+	       jmin = 0;
+	  if (local_x_start + ny == n_last_stored && n_last % 2 == 0)
+	       --n_last_new; /* Nyquist freq. should not be in other half */
+	  n_last_stored = ny;
+     }
+     else { /* rank == 3 */
+	  ny = nx;
+	  nx = d->local_ny;
+	  nxmax = local_x_start ? nx - 1 : nx;
+	  n_other = nx * ny;
+     }
+#  endif /* HAVE_MPI */
+
+     /* compute p = exp(i*phase) factors: */
+     phasex *= -TWOPI; phasey *= -TWOPI; phasez *= -TWOPI;
+     switch (rank) { /* treat z as the last/truncated dimension always */
+	 case 3: break;
+#  if defined(HAVE_MPI) && ! defined(SCALAR_COMPLEX)
+	 case 2: phasez = phasex; phasex = phasey; phasey = 0; break;
+#  else
+	 case 2: phasez = phasey; phasey = 0; break;
+#  endif
+	 case 1: phasez = phasex; phasex = phasey = 0; break;
+     }
+     CASSIGN_SCALAR(pz, cos(phasez), sin(phasez));
+     phasex += phasez;
+     CASSIGN_SCALAR(pxz, cos(phasex), sin(phasex));
+     phasex += phasey;
+     CASSIGN_SCALAR(pxyz, cos(phasex), sin(phasex));
+     phasey += phasez;
+     CASSIGN_SCALAR(pyz, cos(phasey), sin(phasey));
+
+/* convenience macros to copy cscalars, cscalars times phases, 
+   and conjugated cscalars (THIS IS THE ONLY CODE THAT WAS CHANGED
+   COMPARED TO vectorfield_otherhalf): */
+#  define ASSIGN_V(f,k,f2,k2) { f[k] = f2[k2]; }
+#  define ASSIGN_VP(f,k,f2,k2,p) { CASSIGN_MULT(f[k], f2[k2], p); }
+#  define ASSIGN_CV(f,k,f2,k2) { CASSIGN_CONJ(f[k], f2[k2]); }
+
+     /* First, swap the order of elements and multiply by exp(ikR)
+        phase factors.  We have to be careful here not to double-swap
+        any element pair; this is prevented by never swapping with a
+        "conjugated" point that is earlier in the array.  */
+
+     if (rank == 3) {
+	  int ix, iy;
+	  for (ix = 0; 2*ix <= nxmax; ++ix) {
+	       int xdiff, ixc;
+#  ifdef HAVE_MPI
+	       if (local_x_start == 0) {
+		    xdiff = ix != 0; ixc = (nx - ix) % nx;
+	       }
+	       else {
+		    xdiff = 1; ixc = nx-1 - ix;
+	       }
+#  else
+	       xdiff = ix != 0; ixc = (nx - ix) % nx;
+#  endif
+	       for (iy = 0; iy < ny; ++iy) {
+		    int ydiff = iy != 0;
+		    int i = ix * ny + iy, ic = ixc * ny + (ny - iy) % ny, jmax;
+		    if (ic < i)
+			 continue;
+		    jmax = n_last_new;
+		    if (ic == i)
+			 jmax = (jmax + 1) / 2;
+		    for (j = 1; j <= jmax; ++j) {
+			 int jc = n_last_new + 1 - j;
+			 int ij = i*n_last_stored + j;
+			 int ijc = ic*n_last_stored + jc;
+			 scalar_complex f_tmp[3];
+			 switch (xdiff*2 + ydiff) { /* pick exp(-ikR) phase */
+			     case 3: /* xdiff && ydiff */
+				  ASSIGN_VP(f_tmp, 0, field, ijc, pxyz);
+				  ASSIGN_VP(field, ijc, field, ij, pxyz);
+				  ASSIGN_V(field, ij, f_tmp, 0);
+				  break;
+			     case 2: /* xdiff && !ydiff */
+				  ASSIGN_VP(f_tmp, 0, field, ijc, pxz);
+				  ASSIGN_VP(field, ijc, field, ij, pxz);
+				  ASSIGN_V(field, ij, f_tmp, 0);
+				  break;
+			     case 1: /* !xdiff && ydiff */
+				  ASSIGN_VP(f_tmp, 0, field, ijc, pyz);
+				  ASSIGN_VP(field, ijc, field, ij, pyz);
+				  ASSIGN_V(field, ij, f_tmp, 0);
+				  break;
+			     case 0: /* !xdiff && !ydiff */
+				  ASSIGN_VP(f_tmp, 0, field, ijc, pz);
+				  ASSIGN_VP(field, ijc, field, ij, pz);
+				  ASSIGN_V(field, ij, f_tmp, 0);
+				  break;
+			 }
+		    }
+	       }
+	  }
+
+	  /* Next, conjugate, and remove the holes from the array
+	     corresponding to the DC and Nyquist frequencies (which were in
+	     the first half already): */
+	  for (i = 0; i < n_other; ++i)
+	       for (j = 1; j < n_last_new + 1; ++j) {
+		    int ij = i*n_last_stored + j, ijnew = i*n_last_new + j-1;
+		    ASSIGN_CV(field, ijnew, field, ij);
+	       }
+     }
+     else /* if (rank <= 2) */ {
+	  int i;
+	  if (rank == 1) /* (note that 1d MPI transforms are not allowed) */
+	       nx = 1; /* x dimension is handled by j (last dimension) loop */
+
+#  ifdef HAVE_MPI
+	  for (i = 0; i < nx; ++i)
+#  else
+	  for (i = 0; 2*i <= nx; ++i)
+#  endif
+	  {
+	       int xdiff = i != 0, ic = (nx - i) % nx;
+	       int jmax = n_last_new + (jmin - 1);
+#  ifndef HAVE_MPI
+	       if (ic == i)
+		    jmax = (jmax + 1) / 2;
+#  endif
+	       for (j = jmin; j <= jmax; ++j) {
+		    scalar_complex f_tmp[3];
+#  ifdef HAVE_MPI
+		    int jc = jmax + jmin - j;
+		    int ij = j * nx + i;
+		    int ijc = jc * nx + ic;
+		    if (ijc < ij)
+			 break;
+#  else  /* ! HAVE_MPI */
+		    int jc = n_last_new + 1 - j;
+		    int ij = i*n_last_stored + j;
+		    int ijc = ic*n_last_stored + jc;
+#  endif /* ! HAVE_MPI */
+		    if (xdiff) {
+			 ASSIGN_VP(f_tmp, 0, field, ijc, pxz);
+			 ASSIGN_VP(field, ijc, field, ij, pxz);
+			 ASSIGN_V(field, ij, f_tmp, 0);
+		    }
+		    else {
+			 ASSIGN_VP(f_tmp, 0, field, ijc, pz);
+			 ASSIGN_VP(field, ijc, field, ij, pz);
+			 ASSIGN_V(field, ij, f_tmp, 0);
+		    }
+	       }
+	  }
+
+	  /* Next, conjugate, and remove the holes from the array
+	     corresponding to the DC and Nyquist frequencies (which were in
+	     the first half already): */
+	  for (i = 0; i < nx; ++i)
+	       for (j = jmin; j < n_last_new + jmin; ++j) {
+#  ifdef HAVE_MPI
+		    int ij = j*nx + i, ijnew = (j-jmin)*nx + i;
+#  else
+		    int ij = i*n_last_stored + j, ijnew = i*n_last_new + j-1;
+#  endif
+		    ASSIGN_CV(field, ijnew, field, ij);
+	       }
+     }
+
+#  undef ASSIGN_V
+#  undef ASSIGN_VP
+#  undef ASSIGN_CV
+
+#endif /* ! SCALAR_COMPLEX */
+}
+
 /* Similar to vectorfield_otherhalf, above, except that it operates on
    a real scalar field, which is assumed to have come from e.g. the
    absolute values of a complex field (and thus no phase factors or
@@ -876,7 +1191,7 @@ void maxwell_target_operator(evectmatrix Xin, evectmatrix Xout, void *data,
 void maxwell_ucross_op(evectmatrix Xin, evectmatrix Xout,
 		       maxwell_data *d, const real u[3])
 {
-     scalar *fft_data;
+     scalar *fft_data, *fft_data_in;
      scalar_complex *cdata;
      real scale;
      int cur_band_start;
@@ -886,6 +1201,8 @@ void maxwell_ucross_op(evectmatrix Xin, evectmatrix Xout,
      CHECK(Xin.c == 2, "fields don't have 2 components!");
 
      cdata = (scalar_complex *) (fft_data = d->fft_data);
+     fft_data_in = d->fft_data2;
+
      scale = -1.0 / Xout.N;  /* scale factor to normalize FFT;
                                 negative sign comes from 2 i's from curls */
 
@@ -902,7 +1219,7 @@ void maxwell_ucross_op(evectmatrix Xin, evectmatrix Xout,
 		    k_data cur_k = d->k_plus_G[ij];
 		    
 		    for (b = 0; b < cur_num_bands; ++b)
-			 assign_ucross_t2c(&fft_data[3 * (ij2*cur_num_bands
+			 assign_ucross_t2c(&fft_data_in[3 * (ij2*cur_num_bands
 							  + b)], 
 					   u, cur_k, 
 					   &Xin.data[ij * 2 * Xin.p + 
@@ -911,7 +1228,7 @@ void maxwell_ucross_op(evectmatrix Xin, evectmatrix Xout,
 	       }
 	  
 	  /* now, convert to position space via FFT: */
-	  maxwell_compute_fft(+1, d, fft_data,
+	  maxwell_compute_fft(+1, d, fft_data_in, fft_data,
 			      cur_num_bands*3, cur_num_bands*3, 1);
 	  
           maxwell_compute_e_from_d(d, cdata, cur_num_bands);
