@@ -1,4 +1,4 @@
-/* Copyright (C) 1999, 2000, 2001, 2002, Massachusetts Institute of Technology.
+/* Copyright (C) 1999-2014 Massachusetts Institute of Technology.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -23,10 +23,10 @@
 
 #include <ctl.h>
 
-#include "../src/config.h"
+#include "config.h"
 
-#include <check.h>
-#include <matrixio.h>
+#include "check.h"
+#include "matrixio.h"
 
 int verbose = 0;
 
@@ -249,7 +249,8 @@ void map_data(real *d_in_re, real *d_in_im, int n_in[3],
 void handle_dataset(matrixio_id in_file, matrixio_id out_file, 
 		    const char *name_re, const char *name_im,
 		    matrix3x3 Rout, matrix3x3 coord_map,
-		    real *kvector, int resolution, real multiply_size[3],
+		    real *kvector, double resolution, 
+		    scalar_complex scaleby, real multiply_size[3],
 		    int pick_nearest, int transpose)
 {
      real *d_in_re = NULL, *d_in_im = NULL, *d_out_re = NULL, *d_out_im = NULL;
@@ -327,8 +328,18 @@ void handle_dataset(matrixio_id in_file, matrixio_id out_file,
      map_data(d_in_re, d_in_im, in_dims, d_out_re, d_out_im, out_dims,
 	      coord_map, kvector, pick_nearest, transpose);
 
+     if (d_out_im) { /* multiply * scaleby for complex data */
+	  for (i = 0; i < N; ++i) {
+	       scalar_complex d;
+	       CASSIGN_SCALAR(d, d_out_re[i], d_out_im[i]);
+	       CASSIGN_MULT(d, scaleby, d);
+	       d_out_re[i] = CSCALAR_RE(d);
+	       d_out_im[i] = CSCALAR_IM(d);
+	  }
+     }
+
      strcpy(out_name, name_re);
-     if (out_file == in_file)
+     if (out_file.id == in_file.id)
 	  strcat(out_name, "-new");
      if (verbose)
 	  printf("Writing dataset to %s...\n", out_name);
@@ -338,7 +349,7 @@ void handle_dataset(matrixio_id in_file, matrixio_id out_file,
 
      if (d_out_im) {
 	  strcpy(out_name, name_im);
-	  if (out_file == in_file)
+	  if (out_file.id == in_file.id)
 	       strcat(out_name, "-new");
 	  if (verbose)
 	       printf("Writing dataset to %s...\n", out_name);
@@ -358,19 +369,182 @@ void handle_dataset(matrixio_id in_file, matrixio_id out_file,
      free(d_out_im);
 }
 
+void handle_cvector_dataset(matrixio_id in_file, matrixio_id out_file, 
+			    matrix3x3 Rout,
+			    matrix3x3 coord_map,
+			    matrix3x3 cart_map,
+			    real *kvector, 
+			    double resolution, 
+			    scalar_complex scaleby, 
+			    real multiply_size[3],
+			    int pick_nearest, int transpose)
+{
+     real *d_in[3][2] = { {0,0},{0,0},{0,0} };
+     int in_dims[3] = {1,1,1}, out_dims[3] = {1,1,1}, out_dims2[3], rank = 3;
+     int i, N, dim, ri;
+     int start[3] = {0,0,0};
+     matrixio_id data_id;
+
+     for (dim = 0; dim < 3; ++dim)
+	  for (ri = 0; ri < 2; ++ri) {
+	       char nam[] = "x.r";
+	       int dims[3] = {1,1,1}, rnk = 3;
+
+	       nam[0] = 'x' + dim;
+	       nam[2] = ri ? 'i' : 'r';
+	       d_in[dim][ri] 
+		    = matrixio_read_real_data(in_file, nam, &rnk, dims,
+					      0, 0, 0, NULL);
+	       if (!d_in[dim][ri])
+		    goto bad;
+	       if (!dim && !ri) {
+		    rank = rnk;
+		    for (i = 0; i < 3; ++i) in_dims[i] = dims[i];
+	       }
+	       else {
+		    if (rank != rnk || in_dims[0] != dims[0] ||
+			in_dims[1] != dims[1] || in_dims[2] != dims[2])
+			 goto bad;
+	       }
+	  }
+
+     if (verbose)
+	  printf("Found complex vector dataset...\n");
+
+     if (verbose)
+	  printf("Input data is rank %d, size %dx%dx%d.\n",
+		 rank, in_dims[0], in_dims[1], in_dims[2]);
+
+     /* rotate vector field according to cart_map */
+     if (verbose)
+	  printf("Rotating vectors by matrix [ %10f%10f%10f\n"
+		 "                             %10f%10f%10f\n"
+		 "                             %10f%10f%10f ]\n",
+		 cart_map.c0.x, cart_map.c1.x, cart_map.c2.x,
+		 cart_map.c0.y, cart_map.c1.y, cart_map.c2.y,
+		 cart_map.c0.z, cart_map.c1.z, cart_map.c2.z);
+     N = in_dims[0] * in_dims[1] * in_dims[2];
+     for (ri = 0; ri < 2; ++ri)
+	  for (i = 0; i < N; ++i) {
+	       vector3 v;
+	       v.x = d_in[0][ri][i];
+	       v.y = d_in[1][ri][i];
+	       v.z = d_in[2][ri][i];
+	       v = matrix3x3_vector3_mult(cart_map, v);
+	       d_in[0][ri][i] = v.x;
+	       d_in[1][ri][i] = v.y;
+	       d_in[2][ri][i] = v.z;
+	  }
+
+     if (resolution > 0) {
+	  out_dims[0] = vector3_norm(Rout.c0) * resolution + 0.5;
+	  out_dims[1] = vector3_norm(Rout.c1) * resolution + 0.5;
+	  out_dims[2] = vector3_norm(Rout.c2) * resolution + 0.5;
+     }
+     else {
+	  for (i = 0; i < 3; ++i)
+	       out_dims[i] = in_dims[i] * multiply_size[i];
+     }
+     for (i = rank; i < 3; ++i)
+	  out_dims[i] = 1;
+     for (N = 1, i = 0; i < 3; ++i)
+	  N *= (out_dims[i] = MAX2(out_dims[i], 1));
+
+     if (transpose) {
+	  out_dims2[0] = out_dims[1];
+	  out_dims2[1] = out_dims[0];
+	  out_dims2[2] = out_dims[2];
+     }
+     else {
+	  out_dims2[0] = out_dims[0];
+	  out_dims2[1] = out_dims[1];
+	  out_dims2[2] = out_dims[2];
+     }
+
+     if (verbose)
+	  printf("Output data %dx%dx%d.\n",
+		 out_dims2[0], out_dims2[1], out_dims2[2]);
+
+     for (dim = 0; dim < 3; ++dim) {
+	  real *d_out_re, *d_out_im;
+	  char nam[] = "x.r-new";
+
+	  CHK_MALLOC(d_out_re, real, N);
+	  CHK_MALLOC(d_out_im, real, N);
+
+	  map_data(d_in[dim][0], d_in[dim][1], in_dims, 
+		   d_out_re, d_out_im, out_dims,
+		   coord_map, kvector, pick_nearest, transpose);
+
+	  for (i = 0; i < N; ++i) { /* multiply * scaleby */
+	       scalar_complex d;
+	       CASSIGN_SCALAR(d, d_out_re[i], d_out_im[i]);
+	       CASSIGN_MULT(d, scaleby, d);
+	       d_out_re[i] = CSCALAR_RE(d);
+	       d_out_im[i] = CSCALAR_IM(d);
+	  }
+
+	  nam[0] = 'x' + dim;
+	  if (out_file.id != in_file.id)
+	       nam[3] = 0;
+	  if (verbose)
+	       printf("Writing dataset to %s...\n", nam);
+	  data_id = matrixio_create_dataset(out_file, nam,"", rank, out_dims2);
+	  matrixio_write_real_data(data_id, out_dims2, start,1, d_out_re);
+	  matrixio_close_dataset(data_id);
+
+	  nam[2] = 'i';
+	  if (verbose)
+	       printf("Writing dataset to %s...\n", nam);
+	  data_id = matrixio_create_dataset(out_file, nam,"", rank, out_dims2);
+	  matrixio_write_real_data(data_id, out_dims2, start,1, d_out_im);
+	  matrixio_close_dataset(data_id);
+
+	  if (verbose)
+	       printf("Successfully wrote out data.\n");
+     }
+
+     for (dim = 0; dim < 3; ++dim)
+          for (ri = 0; ri < 2; ++ri)
+	       free(d_in[dim][ri]);
+     return;
+
+ bad:
+     for (dim = 0; dim < 3; ++dim)
+          for (ri = 0; ri < 2; ++ri)
+	       free(d_in[dim][ri]);
+     /* try individual datasets */
+     for (dim = 0; dim < 3; ++dim) {
+	  char namr[] = "x.r";
+	  char nami[] = "x.i";
+
+	  namr[0] = 'x' + dim;
+	  nami[0] = 'x' + dim;
+	  handle_dataset(in_file, out_file, namr, nami,
+			 Rout, coord_map, kvector, resolution,
+			 scaleby, multiply_size, pick_nearest, transpose);
+
+	  namr[1] = 0;
+	  handle_dataset(in_file, out_file, namr, NULL,
+			 Rout, coord_map, kvector, resolution,
+			 scaleby, multiply_size, pick_nearest, transpose);
+     }
+}
+
 void handle_file(const char *fname, const char *out_fname,
 		 const char *data_name,
-		 int rectify,  int have_ve, vector3 ve,
-		 int resolution, real multiply_size[3],
+		 int rectify,  int have_ve, vector3 ve, double resolution, 
+		 scalar_complex scaleby, real multiply_size[3],
 		 int pick_nearest, int transpose)
 {
      matrixio_id in_file, out_file;
      real *R, *kvector, *copies;
      int dims[2], rank;
      matrix3x3 Rin = {{1,0,0},{0,1,0},{0,0,1}}, Rout, coord_map;
-#define NUM_DATANAMES 16
+     matrix3x3 cart_map = {{1,0,0},{0,1,0},{0,0,1}};
+#define NUM_DATANAMES 13
      char datanames[NUM_DATANAMES][30] = {
-	  "data", "x", "y", "z",
+	  "data", 
 	  "epsilon.xx", 
 	  "epsilon.xy", 
 	  "epsilon.xz", 
@@ -475,6 +649,20 @@ void handle_file(const char *fname, const char *out_fname,
 				vector3_scale(vector3_dot(Rout.c1, Rout.c2) /
 					      vector3_dot(Rout.c1, Rout.c1),
 					      Rout.c1));
+
+	  cart_map.c0 = unit_vector3(Rout.c0);
+	  cart_map.c1 = unit_vector3(Rout.c1);
+	  cart_map.c2 = unit_vector3(Rout.c2);
+	  cart_map = matrix3x3_inverse(cart_map);
+     }
+
+     if (transpose) { /* swap first two rows of cart_map */
+	  vector3 v;
+	  cart_map = matrix3x3_transpose(cart_map);
+	  v = cart_map.c0;
+	  cart_map.c0 = cart_map.c1;
+	  cart_map.c1 = v;
+	  cart_map = matrix3x3_transpose(cart_map);
      }
 
      Rout.c0 = vector3_scale(multiply_size[0], Rout.c0);
@@ -510,22 +698,27 @@ void handle_file(const char *fname, const char *out_fname,
 	  strcpy(name_re, dname);
 	  handle_dataset(in_file, out_file, name_re, NULL,
 			 Rout, coord_map, kvector, resolution,
-			 multiply_size, pick_nearest, transpose);
+			 scaleby, multiply_size, pick_nearest, transpose);
 
 	  sprintf(name_re, "%s.r", dname);
 	  sprintf(name_im, "%s.i", dname);
 	  handle_dataset(in_file, out_file, name_re, name_im,
 			 Rout, coord_map, kvector, resolution,
-			 multiply_size, pick_nearest, transpose);
+			 scaleby, multiply_size, pick_nearest, transpose);
 
 	  if (data_name)
 	       break;
      }
 
+     /* handle complex vector fields x.{ri}, y.{ri}, z.{ri} */
+     handle_cvector_dataset(in_file, out_file,
+			    Rout, coord_map, cart_map, kvector, resolution,
+			    scaleby, multiply_size, pick_nearest, transpose);
+     
      free(kvector);
 
      matrixio_close(in_file);
-     if (out_file != in_file)
+     if (out_file.id != in_file.id)
 	  matrixio_close(out_file);
 }
 
@@ -543,6 +736,7 @@ void usage(FILE *f)
 	     "    -x <mx>\n"
 	     "    -y <my>\n"
 	     "    -z <mx> : output mx/my/mz periods in the x/y/z directions\n"
+	     " -P <angle> : multiply phase shift of <angle> degrees\n"
 	     "     -m <s> : same as -x <s> -y <s> -z <s>\n"
 	     "         -T : transpose first two dimensions (x & y) of data\n"
 	     "         -p : pixellized output (no grid interpolation)\n"
@@ -582,22 +776,25 @@ else { /* treat as if ":" were at the end of fname */
 int main(int argc, char **argv)
 {
      char *out_fname = NULL, *data_name = NULL;
-     int rectify = 0, resolution = 0, have_ve = 0;
+     int rectify = 0, have_ve = 0;
+     double phaseangle = 0;
+     double resolution = 0;
      vector3 ve = {1,0,0};
      real multiply_size[3] = {1,1,1};
      int pick_nearest = 0, transpose = 0;
      int ifile, c;
      extern char *optarg;
      extern int optind;
+     scalar_complex scaleby = {1,0}, phase;
 
-     while ((c = getopt(argc, argv, "hVvo:x:y:z:m:d:n:prTe:")) != -1)
+     while ((c = getopt(argc, argv, "hVvo:x:y:z:m:d:n:prTe:P:")) != -1)
           switch (c) {
               case 'h':
                    usage(stdout);
                    return EXIT_SUCCESS;
               case 'V':
-                   printf("mpb-data " MPB_VERSION " by Steven G. Johnson.\n"
-"Copyright (C) 1999, 2000, 2001, 2002, Massachusetts Institute of Technology.\n"
+                   printf("mpb-data " PACKAGE_VERSION " by Steven G. Johnson.\n"
+"Copyright (C) 1999-2014 Massachusetts Institute of Technology.\n"
 "This is free software, and you are welcome to redistribute it under the\n"
 "terms of the GNU General Public License (GPL).  mpb-data comes with\n"
 "ABSOLUTELY NO WARRANTY; see the GPL for more details.\n");
@@ -634,10 +831,13 @@ int main(int argc, char **argv)
                    multiply_size[2] = atof(optarg);
                    break;
               case 'n':
-                   resolution = atoi(optarg);
+                   resolution = atof(optarg);
 		   CHECK(resolution > 0,
 			 "invalid resolution for -n (must be positive)");
                    break;
+              case 'P':
+                   phaseangle = atof(optarg);
+		   break;
               case 'p':
                    pick_nearest = 1;
                    break;
@@ -667,6 +867,11 @@ int main(int argc, char **argv)
           usage(stderr);
           return EXIT_FAILURE;
      }
+
+     CASSIGN_SCALAR(phase, 
+		    cos(TWOPI * phaseangle / 360.0),
+		    sin(TWOPI * phaseangle / 360.0));
+     CASSIGN_MULT(scaleby, scaleby, phase);
      
      for (ifile = optind; ifile < argc; ++ifile) {
 	  char *dname, *h5_fname;
@@ -676,7 +881,7 @@ int main(int argc, char **argv)
 
 	  handle_file(h5_fname, out_fname, dname, 
 		      rectify, have_ve, ve, resolution, 
-		      multiply_size, pick_nearest, transpose);
+		      scaleby, multiply_size, pick_nearest, transpose);
 	  
 	  if (out_fname)
                free(out_fname);
